import functools
import hashlib
from collections.abc import Sequence
from typing import cast

from eth_account import Account
from eth_account.messages import SignableMessage, encode_typed_data
from ledgereth.accounts import find_account
from ledgereth.comms import init_dongle
from ledgereth.messages import sign_typed_data_draft
from starknet_crypto_py import get_public_key as rs_get_public_key
from starknet_crypto_py import pedersen_hash as rs_pedersen_hash
from starknet_crypto_py import sign as rs_sign
from starknet_crypto_py import verify as rs_verify
from starknet_py.common import int_from_hex
from starknet_py.constants import EC_ORDER
from starknet_py.utils.typed_data import TypedData, TypedDataDict

from paradex_py.utils import raise_value_error

SHA256_EC_MAX_DIGEST = 2**256


def _padded_hex(x: int) -> str:
    # Hex string should have an even
    # number of characters to convert to bytes.
    hex_str = hex(x)[2:]
    return hex_str if len(hex_str) % 2 == 0 else "0" + hex_str


def _indexed_sha256(seed: int, index: int) -> int:
    digest = hashlib.sha256(bytes.fromhex(_padded_hex(seed) + _padded_hex(index))).hexdigest()
    return int_from_hex(digest)


def _grind_key(key_seed: int, key_value_limit: int) -> int:
    max_allowed_value = SHA256_EC_MAX_DIGEST - (SHA256_EC_MAX_DIGEST % key_value_limit)
    current_index = 0

    key = _indexed_sha256(seed=key_seed, index=current_index)
    while key >= max_allowed_value:
        current_index += 1
        key = _indexed_sha256(seed=key_seed, index=current_index)

    return key % key_value_limit


def _sign_stark_key_message(stark_key_message, l1_private_key: int) -> str:
    encoded = encode_typed_data(full_message=stark_key_message)
    signed = Account.sign_message(encoded, l1_private_key)
    sig_hex = signed.signature.hex()
    # Ensure 0x prefix for compatibility with _get_private_key_from_eth_signature
    # which expects [2:66] to skip "0x" and extract the r value
    return sig_hex if sig_hex.startswith("0x") else "0x" + sig_hex


def _sign_stark_key_message_ledger(message: SignableMessage, eth_account_address: str) -> str:
    dongle = init_dongle()
    account = find_account(eth_account_address, dongle, count=10)
    if account is None:
        return raise_value_error(f"Account {eth_account_address} not found on Ledger")
    # header/body is eth_account naming, presumably to be generic
    domain_hash = message.header
    message_hash = message.body
    signed = sign_typed_data_draft(
        domain_hash=domain_hash,
        message_hash=message_hash,
        sender_path=account.path,
        dongle=dongle,
    )
    return signed.signature


def _get_private_key_from_eth_signature(eth_signature_hex: str) -> int:
    r = eth_signature_hex[2 : 64 + 2]
    return _grind_key(int_from_hex(r), EC_ORDER)


def derive_stark_key(l1_private_key: int, stark_key_msg: TypedDataDict) -> int:
    message_signature = _sign_stark_key_message(stark_key_msg, l1_private_key)
    l2_private_key = _get_private_key_from_eth_signature(message_signature)
    return l2_private_key


def derive_stark_key_from_ledger(eth_account_address: str, stark_key_msg: TypedDataDict) -> int:
    signable_message = encode_typed_data(full_message=stark_key_msg)  # type: ignore[arg-type]
    message_signature = _sign_stark_key_message_ledger(signable_message, eth_account_address)
    l2_private_key = _get_private_key_from_eth_signature(message_signature)
    return l2_private_key


def flatten_signature(sig: list[int]) -> str:
    return f'["{sig[0]}","{sig[1]}"]'


def unflatten_signature(sig: str) -> list:
    return [int(x) for x in sig[2:-2].split('","')]


def typed_data_to_message_hash(typed_data: TypedData | TypedDataDict, address: int) -> int:
    typed_data_dataclass = TypedData.from_dict(cast(TypedDataDict, typed_data))
    return typed_data_dataclass.message_hash(address)


# ###
# Override functions in starknet_py.hash.utils that use cpp
# to use the starknet_crypto_py library
# ###


def private_to_stark_key(priv_key: int) -> int:
    """
    Deduces the public key given a private key.
    """
    return rs_get_public_key(priv_key)


def pedersen_hash(left: int, right: int) -> int:
    """
    One of two hash functions (along with _starknet_keccak) used throughout Starknet.
    """
    return rs_pedersen_hash(left, right)


def compute_hash_on_elements(data: Sequence) -> int:
    """
    Computes a hash chain over the data, in the following order:
        h(h(h(h(0, data[0]), data[1]), ...), data[n-1]), n).

    The hash is initialized with 0 and ends with the data length appended.
    The length is appended in order to avoid collisions of the following kind:
    H([x,y,z]) = h(h(x,y),z) = H([w, z]) where w = h(x,y).
    """
    return functools.reduce(pedersen_hash, [*data, len(data)], 0)


def message_signature(msg_hash: int, priv_key: int, seed: int = 32) -> tuple[int, int]:
    """
    Signs the message with private key.
    """
    # `seed`: extra seed for additional entropy
    # `k`: is generated from `seed` by `starknet-crypto-py`
    # Ref: https://github.com/tradeparadex/starknet-crypto-py/blob/v0.2.0/src/lib.rs#L33
    return rs_sign(private_key=priv_key, msg_hash=msg_hash, seed=seed)


def verify_message_signature(msg_hash: int, signature: list[int], public_key: int) -> bool:
    """
    Verifies ECDSA signature of a given message hash with a given public key.
    Returns true if public_key signs the message.
    """
    r, s = signature
    return rs_verify(msg_hash=msg_hash, r=r, s=s, public_key=public_key)



"""Tests for user_agent module."""

import platform
import sys
from unittest.mock import patch

from paradex_py.user_agent import get_user_agent


class TestUserAgent:
    """Test suite for user agent generation."""

    def test_user_agent_format(self):
        """Test that user agent follows expected format."""
        user_agent = get_user_agent()

        # Should contain all required components
        assert "paradex-py/" in user_agent
        assert "Python" in user_agent
        assert platform.system() in user_agent

    def test_user_agent_includes_sdk_version(self):
        """Test that user agent includes SDK version."""
        user_agent = get_user_agent()

        # Should have version after paradex-py/
        assert user_agent.startswith("paradex-py/")
        version_part = user_agent.split("paradex-py/")[1].split(" ")[0]
        # Version should be either semver format or "dev"
        assert version_part == "dev" or "." in version_part

    def test_user_agent_includes_python_version(self):
        """Test that user agent includes Python version."""
        user_agent = get_user_agent()

        expected_python = f"Python {sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
        assert expected_python in user_agent

    def test_user_agent_includes_platform(self):
        """Test that user agent includes platform."""
        user_agent = get_user_agent()

        platform_name = platform.system()
        assert platform_name in user_agent

    def test_user_agent_fallback_dev_version(self):
        """Test that user agent falls back to 'dev' when package not installed."""
        from importlib.metadata import PackageNotFoundError

        with patch("paradex_py.user_agent.version", side_effect=PackageNotFoundError("Package not found")):
            user_agent = get_user_agent()

            # Should fallback to "dev"
            assert "paradex-py/dev" in user_agent

    def test_user_agent_structure(self):
        """Test that user agent has correct structure."""
        user_agent = get_user_agent()

        # Expected format: "paradex-py/{VERSION} (Python {PYTHON_VERSION}; {PLATFORM})"
        parts = user_agent.split(" ", 1)
        assert len(parts) == 2
        assert parts[0].startswith("paradex-py/")
        assert parts[1].startswith("(")
        assert parts[1].endswith(")")

        # Check inner part
        inner = parts[1][1:-1]  # Remove parentheses
        inner_parts = inner.split("; ")
        assert len(inner_parts) == 2
        assert inner_parts[0].startswith("Python ")
        # inner_parts[1] is the platform

    def test_user_agent_consistent(self):
        """Test that user agent returns consistent value across multiple calls."""
        user_agent1 = get_user_agent()
        user_agent2 = get_user_agent()

        assert user_agent1 == user_agent2



"""Simple WebSocket timeout tests that don't hang."""

from paradex_py import Paradex
from paradex_py.api.ws_client import ParadexWebsocketClient
from paradex_py.constants import WS_TIMEOUT
from paradex_py.environment import TESTNET

MOCK_L1_PRIVATE_KEY = "0x0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef"
MOCK_L1_ADDRESS = "0xabcdef0123456789abcdef0123456789abcdef01"


class TestWebSocketTimeout:
    """Simple test suite for WebSocket timeout configuration functionality."""

    def test_default_timeout_from_constant(self):
        """Test that WebSocket client uses default timeout from constants."""
        ws_client = ParadexWebsocketClient(env=TESTNET, auto_start_reader=False)
        assert ws_client.ws_timeout == WS_TIMEOUT
        assert ws_client.ws_timeout == 20  # Current default value

    def test_custom_timeout_parameter(self):
        """Test that custom timeout parameter is properly set."""
        custom_timeout = 2
        ws_client = ParadexWebsocketClient(env=TESTNET, ws_timeout=custom_timeout, auto_start_reader=False)
        assert ws_client.ws_timeout == custom_timeout

    def test_none_timeout_uses_default(self):
        """Test that passing None for timeout uses the default."""
        ws_client = ParadexWebsocketClient(env=TESTNET, ws_timeout=None, auto_start_reader=False)
        assert ws_client.ws_timeout == WS_TIMEOUT

    def test_timeout_parameter_types(self):
        """Test that timeout parameter accepts different integer values."""
        test_values = [1, 2, 3, 5, 10]
        for timeout_value in test_values:
            ws_client = ParadexWebsocketClient(env=TESTNET, ws_timeout=timeout_value, auto_start_reader=False)
            assert ws_client.ws_timeout == timeout_value

    def test_paradex_class_timeout_passthrough_default(self):
        """Test that Paradex class properly creates WebSocket client with default timeout."""
        paradex = Paradex(env=TESTNET)
        assert paradex.ws_client.ws_timeout == WS_TIMEOUT

    def test_paradex_class_timeout_passthrough_custom(self):
        """Test that Paradex class properly passes custom timeout to WebSocket client."""
        custom_timeout = 3
        paradex = Paradex(env=TESTNET, ws_timeout=custom_timeout)
        assert paradex.ws_client.ws_timeout == custom_timeout

    def test_paradex_class_timeout_none_uses_default(self):
        """Test that passing None for timeout uses the default."""
        paradex = Paradex(env=TESTNET, ws_timeout=None)
        assert paradex.ws_client.ws_timeout == WS_TIMEOUT

    def test_timeout_edge_cases(self):
        """Test edge cases for timeout values."""
        # Test minimum reasonable values
        ws_client = ParadexWebsocketClient(env=TESTNET, ws_timeout=1)
        assert ws_client.ws_timeout == 1

        # Test zero (though not recommended in practice)
        ws_client = ParadexWebsocketClient(env=TESTNET, ws_timeout=0)
        assert ws_client.ws_timeout == 0

        # Test larger values
        ws_client = ParadexWebsocketClient(env=TESTNET, ws_timeout=10)
        assert ws_client.ws_timeout == 10

    def test_timeout_with_account_initialization(self):
        """Test that timeout setting persists through account initialization."""
        custom_timeout = 2
        paradex = Paradex(
            env=TESTNET, l1_address=MOCK_L1_ADDRESS, l1_private_key=MOCK_L1_PRIVATE_KEY, ws_timeout=custom_timeout
        )

        # Ensure timeout is preserved after account initialization
        assert paradex.ws_client.ws_timeout == custom_timeout
        assert paradex.account is not None  # Account should be initialized

    def test_timeout_parameter_independence(self):
        """Test that different WebSocket clients can have different timeouts."""
        timeout1 = 1
        timeout2 = 3

        ws_client1 = ParadexWebsocketClient(env=TESTNET, ws_timeout=timeout1)
        ws_client2 = ParadexWebsocketClient(env=TESTNET, ws_timeout=timeout2)

        assert ws_client1.ws_timeout == timeout1
        assert ws_client2.ws_timeout == timeout2
        assert ws_client1.ws_timeout != ws_client2.ws_timeout

    def test_backward_compatibility(self):
        """Test that existing code without timeout parameter still works."""
        # This should work exactly as before
        ws_client = ParadexWebsocketClient(env=TESTNET, auto_start_reader=False)
        assert ws_client.ws_timeout == WS_TIMEOUT

        # Paradex class without timeout should also work
        paradex = Paradex(env=TESTNET)
        assert paradex.ws_client.ws_timeout == WS_TIMEOUT

    def test_integration_with_existing_features(self):
        """Test that timeout configuration works with other WebSocket client features."""
        custom_timeout = 2
        ws_client = ParadexWebsocketClient(env=TESTNET, ws_timeout=custom_timeout, auto_start_reader=False)

        # Test that other properties are still properly initialized
        assert ws_client.env == TESTNET
        assert ws_client.callbacks == {}
        assert ws_client.subscribed_channels == {}
        assert ws_client.ws_timeout == custom_timeout
        assert ws_client.api_url == f"wss://ws.api.{TESTNET}.paradex.trade/v1"

    def test_timeout_parameter_in_constructor(self):
        """Test that timeout parameter is properly handled in constructor."""
        # Test with explicit None
        ws_client1 = ParadexWebsocketClient(env=TESTNET, ws_timeout=None)
        assert ws_client1.ws_timeout == WS_TIMEOUT

        # Test with custom value
        ws_client2 = ParadexWebsocketClient(env=TESTNET, ws_timeout=5)
        assert ws_client2.ws_timeout == 5

        # Test with zero
        ws_client3 = ParadexWebsocketClient(env=TESTNET, ws_timeout=0)
        assert ws_client3.ws_timeout == 0



"""User agent utility for HTTP and WebSocket clients."""

import platform
import sys
from importlib.metadata import PackageNotFoundError, version


def get_user_agent() -> str:
    """Generate a formatted User-Agent header for paradex-py SDK.

    Format: paradex-py/{VERSION} (Python {PYTHON_VERSION}; {PLATFORM})

    Returns:
        str: Formatted user agent string

    Examples:
        >>> agent = get_user_agent()
        >>> print(agent)
        paradex-py/0.5.4 (Python 3.11.14; Darwin)
    """
    # Get SDK version
    try:
        sdk_version = version("paradex_py")
    except PackageNotFoundError:
        # Fallback for development mode when package not installed
        sdk_version = "dev"

    # Get Python version (e.g., "3.11.14")
    python_version = f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"

    # Get platform name (e.g., "Darwin", "Linux", "Windows")
    platform_name = platform.system()

    return f"paradex-py/{sdk_version} (Python {python_version}; {platform_name})"



"""Tests for authentication and signing enhancements: auth providers, signers, auto_auth, etc."""

import time
from unittest.mock import MagicMock, patch

from paradex_py.api.api_client import ParadexApiClient
from paradex_py.api.protocols import NoOpSigner
from paradex_py.environment import TESTNET


class MockAuthProvider:
    """Mock authentication provider for testing."""

    def __init__(self, token="mock-test-token", should_refresh=False):  # noqa: S107
        self.token = token
        self.should_refresh = should_refresh
        self.get_token_calls = 0
        self.refresh_calls = 0

    def get_token(self):
        self.get_token_calls += 1
        return self.token

    def refresh_if_needed(self):
        self.refresh_calls += 1
        if self.should_refresh:
            self.token = f"refreshed-token-{self.refresh_calls}"
            return self.token
        return self.token if self.token else None


class MockSigner:
    """Mock signer for testing."""

    def __init__(self, should_add_signature=True):
        self.should_add_signature = should_add_signature
        self.signed_orders = []
        self.signed_batches = []

    def sign_order(self, order_data):
        self.signed_orders.append(order_data.copy())
        if self.should_add_signature:
            result = order_data.copy()
            result["signature"] = "mock-signature"
            return result
        return order_data

    def sign_batch(self, orders):
        self.signed_batches.append([order.copy() for order in orders])
        if self.should_add_signature:
            return [{**order, "signature": f"mock-signature-{i}"} for i, order in enumerate(orders)]
        return orders


class TestAuthenticationEnhancements:
    """Test authentication enhancements."""

    @patch.object(ParadexApiClient, "fetch_system_config", return_value=MagicMock())
    def test_auto_auth_enabled_default(self, mock_config):
        """Test auto_auth is enabled by default."""
        client = ParadexApiClient(env=TESTNET)
        assert client.auto_auth is True

    @patch.object(ParadexApiClient, "fetch_system_config", return_value=MagicMock())
    def test_auto_auth_disabled(self, mock_config):
        """Test auto_auth can be disabled."""
        client = ParadexApiClient(env=TESTNET, auto_auth=False)
        assert client.auto_auth is False

    @patch.object(ParadexApiClient, "fetch_system_config", return_value=MagicMock())
    def test_auth_provider_injection(self, mock_config):
        """Test auth provider injection."""
        auth_provider = MockAuthProvider()
        client = ParadexApiClient(env=TESTNET, auth_provider=auth_provider)
        assert client.auth_provider is auth_provider

    @patch.object(ParadexApiClient, "fetch_system_config", return_value=MagicMock())
    def test_set_token_manual(self, mock_config):
        """Test manual token injection."""
        client = ParadexApiClient(env=TESTNET)
        test_token = "mock-manual-jwt-token"  # noqa: S105

        client.set_token(test_token)

        assert client._manual_token == test_token
        assert "Bearer mock-manual-jwt-token" in str(client.client.headers.get("Authorization", ""))

    @patch.object(ParadexApiClient, "fetch_system_config", return_value=MagicMock())
    def test_validate_auth_with_manual_token_auto_auth_disabled(self, mock_config):
        """Test auth validation with manual token and auto_auth disabled."""
        client = ParadexApiClient(env=TESTNET, auto_auth=False)
        client.set_token("mock-manual-token")

        # Should not raise any errors
        client._validate_auth()

        # Should not attempt to refresh since we have manual token and auto_auth is disabled
        assert client._manual_token == "mock-manual-token"  # noqa: S105

    @patch.object(ParadexApiClient, "fetch_system_config", return_value=MagicMock())
    def test_validate_auth_with_auth_provider(self, mock_config):
        """Test auth validation with custom auth provider."""
        auth_provider = MockAuthProvider(token="mock-provider-token", should_refresh=True)  # noqa: S106
        client = ParadexApiClient(env=TESTNET, auth_provider=auth_provider)

        client._validate_auth()

        # Should have called refresh_if_needed
        assert auth_provider.refresh_calls == 1
        assert "Bearer refreshed-token-1" in str(client.client.headers.get("Authorization", ""))

    @patch.object(ParadexApiClient, "fetch_system_config", return_value=MagicMock())
    def test_validate_auth_with_auth_provider_syncs_account_token(self, mock_config):
        """Test that auth_provider token is synced to account.jwt_token for WebSocket compatibility."""
        auth_provider = MockAuthProvider(token="mock-provider-token", should_refresh=True)  # noqa: S106
        client = ParadexApiClient(env=TESTNET, auth_provider=auth_provider)

        # Mock account with set_jwt_token method
        mock_account = MagicMock()
        mock_account.set_jwt_token = MagicMock()
        client.account = mock_account

        client._validate_auth()

        # Should have called refresh_if_needed
        assert auth_provider.refresh_calls == 1
        # Should have synced token to account
        mock_account.set_jwt_token.assert_called_once_with("refreshed-token-1")
        assert "Bearer refreshed-token-1" in str(client.client.headers.get("Authorization", ""))

    @patch.object(ParadexApiClient, "fetch_system_config", return_value=MagicMock())
    def test_validate_auth_without_account_auto_auth_disabled(self, mock_config):
        """Test auth validation without account when auto_auth is disabled."""
        client = ParadexApiClient(env=TESTNET, auto_auth=False)
        client.account = None

        # Should not raise error when auto_auth is disabled
        client._validate_auth()

    @patch.object(ParadexApiClient, "fetch_system_config", return_value=MagicMock())
    def test_validate_auth_token_expiry_auto_auth_enabled(self, mock_config):
        """Test auth validation with expired token and auto_auth enabled."""
        client = ParadexApiClient(env=TESTNET, auto_auth=True)

        # Mock account
        mock_account = MagicMock()
        client.account = mock_account

        # Set auth timestamp to be old (> 4 minutes ago)
        client.auth_timestamp = time.time() - 300  # 5 minutes ago

        with patch.object(client, "auth") as mock_auth:
            client._validate_auth()
            # Should call auth to refresh token
            mock_auth.assert_called_once()

    @patch.object(ParadexApiClient, "fetch_system_config", return_value=MagicMock())
    def test_validate_auth_token_expiry_auto_auth_disabled(self, mock_config):
        """Test auth validation with expired token and auto_auth disabled."""
        client = ParadexApiClient(env=TESTNET, auto_auth=False)

        # Mock account
        mock_account = MagicMock()
        client.account = mock_account

        # Set auth timestamp to be old (> 4 minutes ago)
        client.auth_timestamp = time.time() - 300  # 5 minutes ago

        with patch.object(client, "auth") as mock_auth, patch.object(client.logger, "warning") as mock_warning:
            client._validate_auth()

            # Should not call auth
            mock_auth.assert_not_called()
            # Should log warning
            mock_warning.assert_called_once()

    @patch.object(ParadexApiClient, "fetch_system_config", return_value=MagicMock())
    def test_init_account_with_auto_auth_enabled(self, mock_config):
        """Test account initialization with auto_auth enabled."""
        client = ParadexApiClient(env=TESTNET, auto_auth=True)
        mock_account = MagicMock()

        with patch.object(client, "onboarding") as mock_onboarding, patch.object(client, "auth") as mock_auth:
            client.init_account(mock_account)

            assert client.account is mock_account
            mock_onboarding.assert_called_once()
            mock_auth.assert_called_once()

    @patch.object(ParadexApiClient, "fetch_system_config", return_value=MagicMock())
    def test_init_account_with_auto_auth_disabled(self, mock_config):
        """Test account initialization with auto_auth disabled."""
        client = ParadexApiClient(env=TESTNET, auto_auth=False)
        mock_account = MagicMock()

        with patch.object(client, "onboarding") as mock_onboarding, patch.object(client, "auth") as mock_auth:
            client.init_account(mock_account)

            assert client.account is mock_account
            # Should not call onboarding or auth when auto_auth is disabled
            mock_onboarding.assert_not_called()
            mock_auth.assert_not_called()


class TestSigningEnhancements:
    """Test signing enhancements."""

    @patch.object(ParadexApiClient, "fetch_system_config", return_value=MagicMock())
    def test_signer_injection(self, mock_config):
        """Test signer injection."""
        signer = MockSigner()
        client = ParadexApiClient(env=TESTNET, signer=signer)
        assert client.signer is signer

    @patch.object(ParadexApiClient, "fetch_system_config", return_value=MagicMock())
    def test_submit_order_with_instance_signer(self, mock_config):
        """Test submit_order with instance signer."""
        signer = MockSigner()
        client = ParadexApiClient(env=TESTNET, signer=signer)

        mock_order = MagicMock()
        mock_order.dump_to_dict.return_value = {"symbol": "BTC-USD-PERP", "side": "BUY"}

        with patch.object(client, "_post_authorized", return_value={"success": True}) as mock_post:
            result = client.submit_order(mock_order)

            assert result == {"success": True}
            assert len(signer.signed_orders) == 1
            assert signer.signed_orders[0] == {"symbol": "BTC-USD-PERP", "side": "BUY"}

            # Should have posted signed order
            mock_post.assert_called_once_with(
                path="orders", payload={"symbol": "BTC-USD-PERP", "side": "BUY", "signature": "mock-signature"}
            )

    @patch.object(ParadexApiClient, "fetch_system_config", return_value=MagicMock())
    def test_submit_order_with_parameter_signer(self, mock_config):
        """Test submit_order with parameter signer overriding instance signer."""
        instance_signer = MockSigner()
        param_signer = MockSigner()
        client = ParadexApiClient(env=TESTNET, signer=instance_signer)

        mock_order = MagicMock()
        mock_order.dump_to_dict.return_value = {"symbol": "BTC-USD-PERP", "side": "BUY"}

        with patch.object(client, "_post_authorized", return_value={"success": True}) as mock_post:
            result = client.submit_order(mock_order, signer=param_signer)

            assert result == {"success": True}

            # Parameter signer should be used
            assert len(param_signer.signed_orders) == 1
            assert len(instance_signer.signed_orders) == 0

            mock_post.assert_called_once_with(
                path="orders", payload={"symbol": "BTC-USD-PERP", "side": "BUY", "signature": "mock-signature"}
            )

    @patch.object(ParadexApiClient, "fetch_system_config", return_value=MagicMock())
    def test_submit_order_with_account_fallback(self, mock_config):
        """Test submit_order falls back to account signing."""
        client = ParadexApiClient(env=TESTNET)  # No signer

        mock_account = MagicMock()
        mock_account.sign_order.return_value = "account-signature"
        client.account = mock_account

        mock_order = MagicMock()
        mock_order.dump_to_dict.return_value = {"symbol": "BTC-USD-PERP", "side": "BUY"}

        with patch.object(client, "_post_authorized", return_value={"success": True}) as mock_post:
            result = client.submit_order(mock_order)

            assert result == {"success": True}

            # Should have used account signing
            mock_account.sign_order.assert_called_once_with(mock_order)
            assert mock_order.signature == "account-signature"

            mock_post.assert_called_once_with(path="orders", payload={"symbol": "BTC-USD-PERP", "side": "BUY"})

    @patch.object(ParadexApiClient, "fetch_system_config", return_value=MagicMock())
    def test_submit_orders_batch_with_signer(self, mock_config):
        """Test submit_orders_batch with signer."""
        signer = MockSigner()
        client = ParadexApiClient(env=TESTNET, signer=signer)

        mock_orders = [MagicMock(), MagicMock()]
        mock_orders[0].dump_to_dict.return_value = {"symbol": "BTC-USD-PERP", "side": "BUY"}
        mock_orders[1].dump_to_dict.return_value = {"symbol": "ETH-USD-PERP", "side": "SELL"}

        with patch.object(client, "_post_authorized", return_value={"success": True}) as mock_post:
            result = client.submit_orders_batch(mock_orders)

            assert result == {"success": True}
            assert len(signer.signed_batches) == 1
            assert len(signer.signed_batches[0]) == 2

            # Should have posted signed batch
            expected_payload = [
                {"symbol": "BTC-USD-PERP", "side": "BUY", "signature": "mock-signature-0"},
                {"symbol": "ETH-USD-PERP", "side": "SELL", "signature": "mock-signature-1"},
            ]
            mock_post.assert_called_once_with(path="orders/batch", payload=expected_payload)

    @patch.object(ParadexApiClient, "fetch_system_config", return_value=MagicMock())
    def test_modify_order_with_signer(self, mock_config):
        """Test modify_order with signer."""
        signer = MockSigner()
        client = ParadexApiClient(env=TESTNET, signer=signer)

        mock_order = MagicMock()
        mock_order.dump_to_dict.return_value = {"symbol": "BTC-USD-PERP", "size": "0.2"}

        with patch.object(client, "_put_authorized", return_value={"success": True}) as mock_put:
            result = client.modify_order("order-123", mock_order)

            assert result == {"success": True}
            assert len(signer.signed_orders) == 1

            mock_put.assert_called_once_with(
                path="orders/order-123",
                payload={"symbol": "BTC-USD-PERP", "size": "0.2", "signature": "mock-signature"},
            )

    def test_noop_signer_integration(self):
        """Test NoOpSigner integration for full simulation."""
        signer = NoOpSigner()

        order_data = {"symbol": "BTC-USD-PERP", "side": "BUY", "size": "0.1"}
        result = signer.sign_order(order_data)

        # Should return unchanged
        assert result is order_data
        assert "signature" not in result

        # Test batch signing
        orders = [order_data, {"symbol": "ETH-USD-PERP", "side": "SELL", "size": "1.0"}]
        batch_result = signer.sign_batch(orders)

        assert batch_result is orders
        assert all("signature" not in order for order in batch_result)



# Generated from Paradex API spec version 1.106.0

from __future__ import annotations

from enum import Enum
from typing import Annotated

from pydantic import BaseModel, ConfigDict, Field

from . import responses


class AccountMarginRequest(BaseModel):
    model_config = ConfigDict(
        extra="allow",
        populate_by_name=True,
    )
    leverage: Annotated[int, Field(description="Leverage value (1 up to market's maximum leverage)", examples=[10])]
    margin_type: Annotated[str, Field(description="Margin type (CROSS or ISOLATED)", examples=["CROSS"])]
    on_behalf_of_account: Annotated[
        str | None,
        Field(
            description="Specify sub-account ID to set margin for, if not provided, it will be the main account itself",
            examples=["0x1234567890abcdef"],
        ),
    ] = None


class CancelOrderBatchRequest(BaseModel):
    model_config = ConfigDict(
        extra="allow",
        populate_by_name=True,
    )
    client_order_ids: Annotated[
        list[str] | None,
        Field(description="List of client order IDs to cancel", examples=[['["client-id-1"', '"client-id-2"]']]),
    ] = None
    order_ids: Annotated[
        list[str] | None,
        Field(description="List of order IDs to cancel", examples=[['["order-id-1"', '"order-id-2"]']]),
    ] = None


class CreateSubkey(BaseModel):
    model_config = ConfigDict(
        extra="allow",
        populate_by_name=True,
    )
    key_type: Annotated[str | None, Field(description="Key type to be registered as a subkey.")] = None
    name: Annotated[
        str | None, Field(description="User-friendly name for the subkey.", examples=["My Trading Subkey"])
    ] = None
    public_key: Annotated[
        str | None,
        Field(
            description="Public key to be registered as a subkey.",
            examples=["0x3d9f2b2e5f50c1aade60ca540368cd7490160f41270c192c05729fe35b656a9"],
        ),
    ] = None


class CreateToken(BaseModel):
    model_config = ConfigDict(
        extra="allow",
        populate_by_name=True,
    )
    expiry_duration: Annotated[
        int | None,
        Field(description="Duration in seconds from now until expiration (1 min to 1 year).", examples=[86400]),
    ] = None
    name: Annotated[
        str | None, Field(description="User-friendly name for the JWT token.", examples=["My Long-term Trading Token"])
    ] = None
    token_type: Annotated[str | None, Field(description="Type of token to create.")] = None


class CreateVault(BaseModel):
    model_config = ConfigDict(
        extra="allow",
        populate_by_name=True,
    )
    deposit_tx_signature: Annotated[
        str | None, Field(description="Initial deposit transfer by vault owner", examples=["["])
    ] = None
    description: Annotated[
        str | None, Field(description="Description for the vault", examples=["My vault description"])
    ] = None
    lockup_period: Annotated[
        int | None, Field(description="User's deposits lockup period in days", examples=[1])
    ] = None
    max_tvl: Annotated[
        int | None, Field(description="Max TVL locked by the Vault, if any. 0 for unlimited", examples=[1000000])
    ] = None
    name: Annotated[str | None, Field(description="Unique name for the vault", examples=["MyVault"])] = None
    profit_share: Annotated[
        int | None, Field(description="Vault owner profit share (percentage)", examples=[10])
    ] = None
    public_key: Annotated[
        str | None, Field(description="Public key of vault operator", examples=["0x1234567890abcdef"])
    ] = None


class CreateXPTransferRequest(BaseModel):
    model_config = ConfigDict(
        extra="allow",
        populate_by_name=True,
    )
    amount: Annotated[int, Field(description="Amount of XP to transfer (must be >= 1)", examples=[100], ge=1)]
    is_private: Annotated[
        bool | None,
        Field(
            description=(
                "Whether the transfer is private (not shown on public feeds). Defaults to false if not provided."
            ),
            examples=[False],
        ),
    ] = False
    recipient: Annotated[
        str,
        Field(description="Recipient's Starknet address (hex format)", examples=["0x1234567890abcdef0123456789abcdef"]),
    ]
    season: Annotated[
        str | None, Field(description="Season identifier. Defaults to season2 if not provided.", examples=["season2"])
    ] = "season2"


class ModifyOrderRequest(BaseModel):
    model_config = ConfigDict(
        extra="allow",
        populate_by_name=True,
    )
    id: Annotated[str, Field(description="Order ID to be modified")]
    market: Annotated[str, Field(description="Market for which order is modified", examples=["BTC-USD-PERP"])]
    on_behalf_of_account: Annotated[
        str | None,
        Field(
            description="ID corresponding to the configured isolated margin account.  Only for isolated margin orders",
            examples=["0x1234567890abcdef"],
        ),
    ] = None
    price: Annotated[str, Field(description="Existing or modified price of the order", examples=["29500.12"])]
    side: Annotated[str, Field(description="Existing side of the order", examples=["BUY"])]
    signature: Annotated[
        str, Field(description='Order signature in as a string "[r,s]" signed by account\'s paradex private key')
    ]
    signature_timestamp: Annotated[
        int, Field(description="Unix timestamp in milliseconds of order creation, used for signature verification")
    ]
    size: Annotated[str, Field(description="Existing or modified size of the order", examples=["1.213"])]
    trigger_price: Annotated[
        str | None, Field(description="Existing or modified trigger price of a TPSL order", examples=["29500.12"])
    ] = None
    type: Annotated[str, Field(description="Existing type of the order", examples=["LIMIT"])]


class PriceKind(str, Enum):
    price_kind_last = "last"
    price_kind_mark = "mark"
    price_kind_underlying = "underlying"


class UpdateAccountMaxSlippageRequest(BaseModel):
    model_config = ConfigDict(
        extra="allow",
        populate_by_name=True,
    )
    max_slippage: str


class UpdateAccountProfileRequest(BaseModel):
    model_config = ConfigDict(
        extra="allow",
        populate_by_name=True,
    )
    announcements: Annotated[bool | None, Field(examples=[False])] = None
    fill_sound: Annotated[bool | None, Field(examples=[True])] = None
    fills: Annotated[bool | None, Field(examples=[True])] = None
    is_username_private: Annotated[bool | None, Field(examples=[True])] = None
    max_slippage: str | None = None
    orders: Annotated[bool | None, Field(examples=[True])] = None
    referral_code: Annotated[str | None, Field(examples=["referral_code"])] = None
    referred_by: Annotated[str | None, Field(examples=["referral_code"])] = None
    size_currency_display: str | None = None
    tap_share_rate: Annotated[str | None, Field(examples=["0.2"])] = None
    trading_value_display: str | None = None
    transfers: Annotated[bool | None, Field(examples=[True])] = None
    username: Annotated[str | None, Field(examples=["username"])] = None
    xp_share_rate: Annotated[str | None, Field(examples=["0.2"])] = None


class UpdateNotificationPreferencesRequest(BaseModel):
    model_config = ConfigDict(
        extra="allow",
        populate_by_name=True,
    )
    announcements: Annotated[bool | None, Field(examples=[False])] = None
    fill_sound: Annotated[bool | None, Field(examples=[True])] = None
    fills: Annotated[bool | None, Field(examples=[True])] = None
    orders: Annotated[bool | None, Field(examples=[True])] = None
    transfers: Annotated[bool | None, Field(examples=[True])] = None


class UpdateSizeCurrencyDisplayRequest(BaseModel):
    model_config = ConfigDict(
        extra="allow",
        populate_by_name=True,
    )
    size_currency_display: str


class UpdateTradingValueDisplayRequest(BaseModel):
    model_config = ConfigDict(
        extra="allow",
        populate_by_name=True,
    )
    trading_value_display: str


class Utm(BaseModel):
    model_config = ConfigDict(
        extra="allow",
        populate_by_name=True,
    )
    campaign: Annotated[str | None, Field(description="UTM campaign parameter", examples=["summer2024"])] = None
    source: Annotated[str | None, Field(description="UTM source parameter", examples=["google"])] = None
    type: Annotated[str | None, Field(description="UTM type parameter", examples=["clickthrough"])] = None


class AlgoOrderRequest(BaseModel):
    model_config = ConfigDict(
        extra="allow",
        populate_by_name=True,
    )
    algo_type: Annotated[str, Field(description="Algo type, required for algo orders creation", examples=["TWAP"])]
    duration_seconds: Annotated[
        int,
        Field(
            description=(
                "Duration in seconds for which the algo order will be running, required for algo orders creation"
            ),
            examples=[3600],
        ),
    ]
    market: Annotated[str, Field(description="Market for which order is created", examples=["BTC-USD-PERP"])]
    on_behalf_of_account: Annotated[
        str | None,
        Field(
            description="ID corresponding to the configured isolated margin account. Only for isolated margin orders",
            examples=["0x1234567890abcdef"],
        ),
    ] = None
    recv_window: Annotated[
        int | None,
        Field(
            description=(
                "Order will be created if it is received by API within RecvWindow milliseconds from signature"
                " timestamp, minimum is 10 milliseconds"
            )
        ),
    ] = None
    side: Annotated[responses.OrderSide, Field(description="Algo order side", examples=["MARKET"])]
    signature: Annotated[
        str, Field(description='Order signature in as a string "[r,s]" signed by account\'s paradex private key')
    ]
    signature_timestamp: Annotated[
        int, Field(description="Unix timestamp in milliseconds of order creation, used for signature verification")
    ]
    size: Annotated[str, Field(description="Size of the algo order", examples=["1.213"])]
    type: Annotated[responses.OrderType, Field(description="Algo order type, only MARKET is supported")]


class BlockExecuteRequest(BaseModel):
    model_config = ConfigDict(
        extra="allow",
        populate_by_name=True,
    )
    execution_nonce: Annotated[
        str, Field(description="Unique nonce for this execution request", examples=["execution_001"])
    ]
    selected_offers: Annotated[
        list[str] | None,
        Field(description="Array of offer IDs selected for execution (offers are atomic, not partial)"),
    ] = None
    signatures: Annotated[
        dict[str, responses.BlockTradeSignature],
        Field(
            description=(
                "Map of offer IDs to initiator signatures accepting each offer. Block id if it is a direct block trade."
            )
        ),
    ]


class Onboarding(BaseModel):
    model_config = ConfigDict(
        extra="allow",
        populate_by_name=True,
    )
    marketing_code: Annotated[
        str | None,
        Field(description="Marketing code of the campaign the user is being onboarded via.", examples=["abcd:code1"]),
    ] = None
    public_key: Annotated[
        str | None,
        Field(
            description="Public key of the user being onboarded.",
            examples=["0x3d9f2b2e5f50c1aade60ca540368cd7490160f41270c192c05729fe35b656a9"],
        ),
    ] = None
    referral_code: Annotated[
        str | None,
        Field(description="Referral code of the user who referred the user being onboarded.", examples=["cryptofox8"]),
    ] = None
    utm: Annotated[Utm | None, Field(description="UTM tracking parameters (optional).")] = None


class OrderRequest(BaseModel):
    model_config = ConfigDict(
        extra="allow",
        populate_by_name=True,
    )
    client_id: Annotated[
        str | None, Field(description="Unique client assigned ID for the order", examples=["123454321"], max_length=64)
    ] = None
    flags: Annotated[list[responses.OrderFlag] | None, Field(description="Order flags, allow flag: REDUCE_ONLY")] = None
    instruction: Annotated[
        responses.OrderInstruction, Field(description="Order Instruction, GTC, IOC, RPI or POST_ONLY if empty GTC")
    ]
    market: Annotated[str, Field(description="Market for which order is created", examples=["BTC-USD-PERP"])]
    on_behalf_of_account: Annotated[
        str | None,
        Field(
            description="ID corresponding to the configured isolated margin account. Only for isolated margin orders",
            examples=["0x1234567890abcdef"],
        ),
    ] = None
    price: Annotated[str, Field(description="Order price", examples=["29500.12"])]
    recv_window: Annotated[
        int | None,
        Field(
            description=(
                "Order will be created if it is received by API within RecvWindow milliseconds from signature"
                " timestamp, minimum is 10 milliseconds"
            )
        ),
    ] = None
    side: Annotated[responses.OrderSide, Field(description="Order side")]
    signature: Annotated[
        str, Field(description='Order signature in as a string "[r,s]" signed by account\'s paradex private key')
    ]
    signature_timestamp: Annotated[
        int, Field(description="Unix timestamp in milliseconds of order creation, used for signature verification")
    ]
    size: Annotated[str, Field(description="Size of the order", examples=["1.213"])]
    stp: Annotated[
        str | None,
        Field(description="Self Trade Prevention, EXPIRE_MAKER, EXPIRE_TAKER or EXPIRE_BOTH, if empty EXPIRE_TAKER"),
    ] = None
    trigger_price: Annotated[str | None, Field(description="Trigger price for stop order")] = None
    type: Annotated[responses.OrderType, Field(description="Order type")]
    vwap_price: Annotated[
        str | None,
        Field(description="Optional VWAP price for market orders used for VWAP instruction, bounded by price band"),
    ] = None


class BlockOfferInfo(BaseModel):
    model_config = ConfigDict(
        extra="allow",
        populate_by_name=True,
    )
    offerer_order: Annotated[responses.BlockTradeOrder, Field(description="Order details from the offering account")]
    price: Annotated[str, Field(description="Offered price for this market", examples=["30050.00"])]
    size: Annotated[str, Field(description="Offered size for this market", examples=["5.0"])]


class BlockOfferRequest(BaseModel):
    model_config = ConfigDict(
        extra="allow",
        populate_by_name=True,
    )
    nonce: Annotated[str, Field(description="Unique nonce for this offer request", examples=["98765"])]
    offering_account: Annotated[
        str, Field(description="Starknet address of the account making this offer", examples=["0xabcdef1234567890"])
    ]
    signature: Annotated[
        responses.BlockTradeSignature, Field(description="Cryptographic signature authorizing this offer")
    ]
    trades: Annotated[dict[str, BlockOfferInfo], Field(description="Map of market symbol to offer details")]


class BlockTradeInfo(BaseModel):
    model_config = ConfigDict(
        extra="allow",
        populate_by_name=True,
    )
    maker_order: Annotated[
        responses.BlockTradeOrder | None, Field(description="Maker order details (empty if requiring offers)")
    ] = None
    price: Annotated[
        str | None, Field(description="Agreed price for this trade (empty if requiring offers)", examples=["30000.00"])
    ] = None
    size: Annotated[
        str | None, Field(description="Agreed size for this trade (empty if requiring offers)", examples=["10.5"])
    ] = None
    taker_order: Annotated[
        responses.BlockTradeOrder | None, Field(description="Taker order details (empty if requiring offers)")
    ] = None
    trade_constraints: Annotated[
        responses.BlockTradeConstraints | None, Field(description="Constraints for this trade when requiring offers")
    ] = None


class BlockTradeRequest(BaseModel):
    model_config = ConfigDict(
        extra="allow",
        populate_by_name=True,
    )
    block_expiration: Annotated[
        int, Field(description="Unix timestamp in milliseconds when block expires", examples=[1640995800000])
    ]
    nonce: Annotated[str, Field(description="Unique nonce for this block trade request", examples=["67890"])]
    required_signers: Annotated[
        list[str], Field(description="List of accounts that can participate in the block trade")
    ]
    signatures: Annotated[
        dict[str, responses.BlockTradeSignature],
        Field(description="Map of account addresses to their signatures. Can be empty or partial."),
    ]
    trades: Annotated[
        dict[str, BlockTradeInfo], Field(description="Map of market symbol to trade info (one per market)")
    ]



"""Tests for ParadexAccount httpx integration."""

from types import MethodType
from unittest.mock import AsyncMock, Mock, patch

import httpx
import pytest
from starknet_py.net.http_client import HttpMethod

from paradex_py.account.account import ParadexAccount
from tests.mocks.api_client import MockApiClient


class TestAccountHttpxIntegration:
    """Test suite for ParadexAccount's httpx integration via monkey patching."""

    def setup_method(self):
        """Setup method to create test fixtures."""
        self.api_client = MockApiClient()
        self.config = self.api_client.fetch_system_config()

        # Test account data
        self.test_l1_address = "0xd2c7314539dCe7752c8120af4eC2AA750Cf2035e"
        self.test_l1_private_key = "0xf8e4d1d772cdd44e5e77615ad11cc071c94e4c06dc21150d903f28e6aa6abdff"

    def test_account_initialization_with_httpx(self):
        """Test that account initialization properly sets up httpx integration."""
        account = ParadexAccount(
            config=self.config,
            l1_address=self.test_l1_address,
            l1_private_key=self.test_l1_private_key,
        )

        # Verify account was created successfully
        assert account.l1_address == self.test_l1_address
        assert account.starknet is not None

        # Verify that the monkey patch was applied
        original_method = account.starknet.client._client._make_request
        assert isinstance(original_method, MethodType)

    @pytest.mark.asyncio
    async def test_monkey_patched_make_request_success(self):
        """Test successful HTTP request through monkey-patched method."""
        account = ParadexAccount(
            config=self.config,
            l1_address=self.test_l1_address,
            l1_private_key=self.test_l1_private_key,
        )

        # Mock httpx.AsyncClient response
        mock_response = Mock()
        mock_response.json = AsyncMock(return_value={"result": "success"})

        # Mock the handle_request_error method to avoid issues
        with patch.object(
            account.starknet.client._client, "handle_request_error", new_callable=AsyncMock
        ) as mock_handle_error:
            mock_handle_error.return_value = None

            # Mock the AsyncClient.request method
            mock_async_client = AsyncMock()
            mock_async_client.request.return_value = mock_response

            # Test the monkey-patched method directly
            payload = {"method": "test", "params": []}
            result = await account.starknet.client._client._make_request(
                session=mock_async_client,
                address="https://test.example.com",
                http_method=HttpMethod.POST,
                params={},
                payload=payload,
            )

            assert result == {"result": "success"}

            # Verify the request was made with correct parameters
            mock_async_client.request.assert_called_once()
            call_args = mock_async_client.request.call_args

            assert call_args[1]["method"] == "POST"
            assert call_args[1]["url"] == "https://test.example.com"
            assert call_args[1]["json"] == payload
            assert "PARADEX-STARKNET-ACCOUNT" in call_args[1]["headers"]
            assert "PARADEX-STARKNET-SIGNATURE" in call_args[1]["headers"]

    @pytest.mark.asyncio
    async def test_monkey_patched_request_with_headers(self):
        """Test that custom headers are properly added to requests."""
        account = ParadexAccount(
            config=self.config,
            l1_address=self.test_l1_address,
            l1_private_key=self.test_l1_private_key,
        )

        # Mock httpx response
        mock_response = Mock()
        mock_response.json = AsyncMock(return_value={"status": "ok"})

        # Mock the handle_request_error method
        with patch.object(
            account.starknet.client._client, "handle_request_error", new_callable=AsyncMock
        ) as mock_handle_error:
            mock_handle_error.return_value = None

            mock_async_client = AsyncMock()
            mock_async_client.request.return_value = mock_response

            payload = {"test": "data"}

            # Call the monkey-patched method
            await account.starknet.client._client._make_request(
                session=mock_async_client,
                address="https://fullnode.example.com/rpc",
                http_method=HttpMethod.POST,
                params={},
                payload=payload,
            )

        # Verify headers were added
        call_args = mock_async_client.request.call_args
        headers = call_args[1]["headers"]

        # Check that all required Paradex headers are present
        required_headers = [
            "PARADEX-STARKNET-ACCOUNT",
            "PARADEX-STARKNET-SIGNATURE",
            "PARADEX-STARKNET-SIGNATURE-TIMESTAMP",
            "PARADEX-STARKNET-SIGNATURE-VERSION",
        ]

        for header in required_headers:
            assert header in headers, f"Missing required header: {header}"

        # Verify account address format
        assert headers["PARADEX-STARKNET-ACCOUNT"].startswith("0x")

        # Verify signature format (should be JSON array string)
        signature = headers["PARADEX-STARKNET-SIGNATURE"]
        assert signature.startswith("[") and signature.endswith("]")

        # Verify timestamp is numeric string
        timestamp = headers["PARADEX-STARKNET-SIGNATURE-TIMESTAMP"]
        assert timestamp.isdigit()

    def test_fullnode_request_headers_generation(self):
        """Test fullnode request headers generation."""
        account = ParadexAccount(
            config=self.config,
            l1_address=self.test_l1_address,
            l1_private_key=self.test_l1_private_key,
        )

        json_payload = '{"method": "test", "params": []}'
        headers = account.fullnode_request_headers(account.starknet, account.l2_chain_id, json_payload)

        # Verify all required headers are present
        assert "Content-Type" in headers
        assert headers["Content-Type"] == "application/json"
        assert "PARADEX-STARKNET-ACCOUNT" in headers
        assert "PARADEX-STARKNET-SIGNATURE" in headers
        assert "PARADEX-STARKNET-SIGNATURE-TIMESTAMP" in headers
        assert "PARADEX-STARKNET-SIGNATURE-VERSION" in headers

        # Verify account address format
        assert headers["PARADEX-STARKNET-ACCOUNT"] == hex(account.starknet.address)

        # Verify signature version
        assert headers["PARADEX-STARKNET-SIGNATURE-VERSION"] == "1.0.0"

    @pytest.mark.asyncio
    async def test_error_handling_in_monkey_patch(self):
        """Test error handling in the monkey-patched method."""
        account = ParadexAccount(
            config=self.config,
            l1_address=self.test_l1_address,
            l1_private_key=self.test_l1_private_key,
        )

        # Mock httpx client that raises an exception
        mock_async_client = AsyncMock()
        mock_response = Mock()

        # Mock handle_request_error to raise the exception
        async def mock_handle_error(response):
            raise httpx.HTTPStatusError("Client error", request=Mock(), response=Mock())

        with patch.object(account.starknet.client._client, "handle_request_error", side_effect=mock_handle_error):
            mock_async_client.request.return_value = mock_response

            # Test that the exception is properly propagated
            with pytest.raises(httpx.HTTPStatusError):
                await account.starknet.client._client._make_request(
                    session=mock_async_client,
                    address="https://test.example.com",
                    http_method=HttpMethod.POST,
                    params={},
                    payload={"test": "data"},
                )

    def test_httpx_client_type_annotation(self):
        """Test that the monkey-patched method has correct type annotation."""
        account = ParadexAccount(
            config=self.config,
            l1_address=self.test_l1_address,
            l1_private_key=self.test_l1_private_key,
        )

        # Get the monkey-patched method
        patched_method = account.starknet.client._client._make_request

        # Verify it's a bound method
        assert isinstance(patched_method, MethodType)

        # The method should accept AsyncClient (httpx) instead of ClientSession (aiohttp)
        # This is verified by the successful execution of other tests

    @pytest.mark.asyncio
    async def test_json_payload_handling(self):
        """Test that JSON payload is properly handled in the monkey-patched method."""
        account = ParadexAccount(
            config=self.config,
            l1_address=self.test_l1_address,
            l1_private_key=self.test_l1_private_key,
        )

        mock_response = Mock()
        mock_response.json = AsyncMock(return_value={"success": True})

        # Mock handle_request_error
        with patch.object(
            account.starknet.client._client, "handle_request_error", new_callable=AsyncMock
        ) as mock_handle_error:
            mock_handle_error.return_value = None

            mock_async_client = AsyncMock()
            mock_async_client.request.return_value = mock_response

            test_payload = {
                "method": "starknet_call",
                "params": ["latest", {"contract_address": "0x123", "entry_point_selector": "0x456"}],
            }

            result = await account.starknet.client._client._make_request(
                session=mock_async_client,
                address="https://fullnode.test.com",
                http_method=HttpMethod.POST,
                params={},
                payload=test_payload,
            )

        # Verify the result
        assert result == {"success": True}

        # Verify the payload was passed correctly
        call_args = mock_async_client.request.call_args
        assert call_args[1]["json"] == test_payload

        # Verify JSON string was created for signature
        # (This is done internally in the method for header generation)
        mock_async_client.request.assert_called_once()



"""Tests for HTTP client enhancements: retry strategies, request hooks, timeouts, etc."""

from unittest.mock import patch

import httpx
import pytest

from paradex_py.api.http_client import HttpClient, HttpMethod
from paradex_py.api.protocols import DefaultRetryStrategy, NoOpSigner


class MockRetryStrategy:
    """Mock retry strategy for testing."""

    def __init__(self, max_retries=2, should_retry_responses=None, should_retry_exceptions=None):
        self.max_retries = max_retries
        self.should_retry_responses = should_retry_responses or []
        self.should_retry_exceptions = should_retry_exceptions or []
        self.delays = []

    def should_retry(self, attempt, response, exception):
        if attempt >= self.max_retries:
            return False
        if response and response.status_code in self.should_retry_responses:
            return True
        return exception and any(isinstance(exception, exc_type) for exc_type in self.should_retry_exceptions)

    def get_delay(self, attempt):
        delay = 0.01 * (2**attempt)  # Very short delays for testing
        self.delays.append(delay)
        return delay


class MockRequestHook:
    """Mock request hook for testing."""

    def __init__(self):
        self.requests = []
        self.responses = []

    def on_request(self, method, url, headers):
        self.requests.append({"method": method, "url": url, "headers": headers})

    def on_response(self, method, url, status_code, duration_ms):
        self.responses.append({"method": method, "url": url, "status_code": status_code, "duration_ms": duration_ms})


class TestHttpClientEnhancements:
    """Test HTTP client enhancements."""

    def setup_method(self):
        """Set up test fixtures."""
        self.http_client = HttpClient()

    def test_default_timeout_configuration(self):
        """Test default timeout configuration."""
        client = HttpClient(default_timeout=30.0)
        assert client.default_timeout == 30.0

        # Test with no default timeout
        client_no_timeout = HttpClient()
        assert client_no_timeout.default_timeout is None

    def test_retry_strategy_configuration(self):
        """Test retry strategy configuration."""
        retry_strategy = MockRetryStrategy(max_retries=3)
        client = HttpClient(retry_strategy=retry_strategy)
        assert client.retry_strategy is retry_strategy

    def test_request_hook_configuration(self):
        """Test request hook configuration."""
        request_hook = MockRequestHook()
        client = HttpClient(request_hook=request_hook)
        assert client.request_hook is request_hook

    @patch("httpx.Client.request")
    def test_request_with_timeout(self, mock_request):
        """Test request with timeout parameter."""
        mock_request.return_value = httpx.Response(200, json={"success": True})

        # Test with per-call timeout
        result = self.http_client.request(url="https://example.com/test", http_method=HttpMethod.GET, timeout=5.0)

        assert result == {"success": True}
        mock_request.assert_called_once()
        call_kwargs = mock_request.call_args[1]
        assert call_kwargs["timeout"] == 5.0

    @patch("httpx.Client.request")
    def test_request_with_default_timeout(self, mock_request):
        """Test request with default timeout."""
        mock_request.return_value = httpx.Response(200, json={"success": True})
        client = HttpClient(default_timeout=10.0)

        result = client.request(url="https://example.com/test", http_method=HttpMethod.GET)

        assert result == {"success": True}
        mock_request.assert_called_once()
        call_kwargs = mock_request.call_args[1]
        assert call_kwargs["timeout"] == 10.0

    @patch("httpx.Client.request")
    def test_request_timeout_override(self, mock_request):
        """Test per-call timeout overrides default timeout."""
        mock_request.return_value = httpx.Response(200, json={"success": True})
        client = HttpClient(default_timeout=10.0)

        result = client.request(url="https://example.com/test", http_method=HttpMethod.GET, timeout=5.0)

        assert result == {"success": True}
        mock_request.assert_called_once()
        call_kwargs = mock_request.call_args[1]
        assert call_kwargs["timeout"] == 5.0

    @patch("httpx.Client.request")
    @patch("time.sleep")  # Mock sleep to speed up tests
    def test_retry_strategy_on_server_error(self, mock_sleep, mock_request):
        """Test retry strategy on server errors."""
        retry_strategy = MockRetryStrategy(max_retries=2, should_retry_responses=[500])
        client = HttpClient(retry_strategy=retry_strategy)

        # All calls return 500, should exhaust retries and raise error
        mock_request.side_effect = [
            httpx.Response(500, json={"error": "INTERNAL_ERROR", "message": "server error", "data": None}),
            httpx.Response(500, json={"error": "INTERNAL_ERROR", "message": "server error", "data": None}),
            httpx.Response(500, json={"error": "INTERNAL_ERROR", "message": "server error", "data": None}),
        ]

        with pytest.raises(ValueError, match="server error"):
            client.request(url="https://example.com/test", http_method=HttpMethod.GET)

        # Should have called request 3 times (original + 2 retries)
        assert mock_request.call_count == 3
        assert mock_sleep.call_count == 2
        assert len(retry_strategy.delays) == 2

    @patch("httpx.Client.request")
    @patch("time.sleep")  # Mock sleep to speed up tests
    def test_retry_strategy_on_exception(self, mock_sleep, mock_request):
        """Test retry strategy on exceptions."""
        retry_strategy = MockRetryStrategy(max_retries=2, should_retry_exceptions=[httpx.RequestError])
        client = HttpClient(retry_strategy=retry_strategy)

        # First call raises exception, second call succeeds
        mock_request.side_effect = [
            httpx.RequestError("Connection failed"),
            httpx.Response(200, json={"success": True}),
        ]

        result = client.request(url="https://example.com/test", http_method=HttpMethod.GET)

        assert result == {"success": True}
        assert mock_request.call_count == 2
        assert mock_sleep.call_count == 1
        assert len(retry_strategy.delays) == 1

    @patch("httpx.Client.request")
    def test_request_hook_called(self, mock_request):
        """Test request hook is called properly."""
        mock_request.return_value = httpx.Response(200, json={"success": True})
        request_hook = MockRequestHook()
        client = HttpClient(request_hook=request_hook)

        result = client.request(
            url="https://example.com/test",
            http_method=HttpMethod.POST,
            headers={"Authorization": "Bearer token"},
        )

        assert result == {"success": True}

        # Check request hook was called
        assert len(request_hook.requests) == 1
        assert len(request_hook.responses) == 1

        request_data = request_hook.requests[0]
        assert request_data["method"] == "POST"
        assert request_data["url"] == "https://example.com/test"
        assert request_data["headers"]["Authorization"] == "[REDACTED]"  # Should be redacted

        response_data = request_hook.responses[0]
        assert response_data["method"] == "POST"
        assert response_data["url"] == "https://example.com/test"
        assert response_data["status_code"] == 200
        assert response_data["duration_ms"] > 0

    def test_header_redaction(self):
        """Test sensitive header redaction."""
        client = HttpClient()
        sensitive_headers = {
            "Authorization": "Bearer secret-token",
            "X-API-Key": "secret-key",
            "JWT": "jwt-token",
            "Token": "auth-token",
            "Content-Type": "application/json",
        }

        redacted = client._redact_headers(sensitive_headers)

        assert redacted["Authorization"] == "[REDACTED]"
        assert redacted["X-API-Key"] == "[REDACTED]"
        assert redacted["JWT"] == "[REDACTED]"
        assert redacted["Token"] == "[REDACTED]"  # noqa: S105
        assert redacted["Content-Type"] == "application/json"  # Not redacted

    def test_header_redaction_case_insensitive(self):
        """Test header redaction is case insensitive."""
        client = HttpClient()
        headers = {"authorization": "Bearer token", "X-Api-Key": "secret"}

        redacted = client._redact_headers(headers)

        assert redacted["authorization"] == "[REDACTED]"
        assert redacted["X-Api-Key"] == "[REDACTED]"

    def test_header_redaction_with_none(self):
        """Test header redaction with None input."""
        client = HttpClient()
        assert client._redact_headers(None) == {}
        assert client._redact_headers({}) == {}


class TestDefaultRetryStrategy:
    """Test the default retry strategy implementation."""

    def test_default_retry_strategy_max_retries(self):
        """Test default retry strategy respects max retries."""
        strategy = DefaultRetryStrategy(max_retries=3)

        # Should retry up to max_retries
        for attempt in range(3):
            assert strategy.should_retry(attempt, httpx.Response(500), None) is True

        # Should not retry beyond max_retries
        assert strategy.should_retry(3, httpx.Response(500), None) is False

    def test_default_retry_strategy_server_errors(self):
        """Test default retry strategy retries on server errors."""
        strategy = DefaultRetryStrategy()

        # Should retry on 5xx errors
        assert strategy.should_retry(0, httpx.Response(500), None) is True
        assert strategy.should_retry(0, httpx.Response(502), None) is True
        assert strategy.should_retry(0, httpx.Response(503), None) is True

        # Should not retry on 4xx errors
        assert strategy.should_retry(0, httpx.Response(400), None) is False
        assert strategy.should_retry(0, httpx.Response(404), None) is False

        # Should not retry on 2xx success
        assert strategy.should_retry(0, httpx.Response(200), None) is False

    def test_default_retry_strategy_rate_limit(self):
        """Test default retry strategy retries on rate limits."""
        strategy = DefaultRetryStrategy()

        # Should retry on 429 rate limit
        assert strategy.should_retry(0, httpx.Response(429), None) is True

    def test_default_retry_strategy_exceptions(self):
        """Test default retry strategy retries on exceptions."""
        strategy = DefaultRetryStrategy()

        # Should retry on network exceptions
        assert strategy.should_retry(0, None, httpx.RequestError("Connection failed")) is True
        assert strategy.should_retry(0, None, Exception("Generic error")) is True

    def test_default_retry_strategy_exponential_backoff(self):
        """Test default retry strategy exponential backoff."""
        strategy = DefaultRetryStrategy(base_delay=1.0, max_delay=10.0)

        # Should implement exponential backoff
        assert strategy.get_delay(0) == 1.0  # 1.0 * 2^0
        assert strategy.get_delay(1) == 2.0  # 1.0 * 2^1
        assert strategy.get_delay(2) == 4.0  # 1.0 * 2^2
        assert strategy.get_delay(3) == 8.0  # 1.0 * 2^3
        assert strategy.get_delay(4) == 10.0  # Capped at max_delay

    def test_default_retry_strategy_custom_parameters(self):
        """Test default retry strategy with custom parameters."""
        strategy = DefaultRetryStrategy(max_retries=5, base_delay=0.5, max_delay=30.0)

        assert strategy.max_retries == 5
        assert strategy.base_delay == 0.5
        assert strategy.max_delay == 30.0

        # Test delay calculation with custom base
        assert strategy.get_delay(0) == 0.5
        assert strategy.get_delay(1) == 1.0
        assert strategy.get_delay(2) == 2.0


class TestNoOpSigner:
    """Test the no-op signer for simulation."""

    def test_noop_signer_sign_order(self):
        """Test no-op signer returns order data unchanged."""
        signer = NoOpSigner()
        order_data = {"symbol": "BTC-USD-PERP", "side": "BUY", "size": "0.1"}

        result = signer.sign_order(order_data)

        assert result is order_data  # Should return same object
        assert result == {"symbol": "BTC-USD-PERP", "side": "BUY", "size": "0.1"}

    def test_noop_signer_sign_batch(self):
        """Test no-op signer returns batch unchanged."""
        signer = NoOpSigner()
        orders = [
            {"symbol": "BTC-USD-PERP", "side": "BUY", "size": "0.1"},
            {"symbol": "ETH-USD-PERP", "side": "SELL", "size": "1.0"},
        ]

        result = signer.sign_batch(orders)

        assert result is orders  # Should return same object
        assert len(result) == 2
        assert result[0] == {"symbol": "BTC-USD-PERP", "side": "BUY", "size": "0.1"}
        assert result[1] == {"symbol": "ETH-USD-PERP", "side": "SELL", "size": "1.0"}



import dataclasses
import logging
import re
from collections.abc import Callable

import marshmallow_dataclass
from starknet_py.constants import RPC_CONTRACT_ERROR
from starknet_py.contract import Contract, InvokeResult
from starknet_py.hash.selector import get_selector_from_name
from starknet_py.net.account.account import Account as StarknetAccount
from starknet_py.net.client import Client
from starknet_py.net.client_errors import ClientError
from starknet_py.net.client_models import Call, Calls, ResourceBoundsMapping, SentTransactionResponse
from starknet_py.net.models import Address, AddressRepresentation, InvokeV3, StarknetChainId
from starknet_py.net.signer import BaseSigner
from starknet_py.net.signer.stark_curve_signer import KeyPair
from starknet_py.proxy.contract_abi_resolver import ProxyConfig
from starknet_py.proxy.proxy_check import ArgentProxyCheck, OpenZeppelinProxyCheck, ProxyCheck
from starknet_py.utils.typed_data import TypedData, TypedDataDict

from .utils import message_signature, typed_data_to_message_hash


class Account(StarknetAccount):
    def __init__(
        self,
        *,
        address: AddressRepresentation,
        client: Client,
        signer: BaseSigner | None = None,
        key_pair: KeyPair | None = None,
        chain: StarknetChainId | None = None,
    ):
        super().__init__(address=address, client=client, signer=signer, key_pair=key_pair, chain=chain)

    def _add_signature(self, invoke: InvokeV3, signature: list[int]) -> InvokeV3:
        return dataclasses.replace(invoke, signature=signature)

    async def prepare_invoke(
        self,
        calls: Calls,
        resource_bounds: ResourceBoundsMapping | None = None,
        auto_estimate: bool = False,
        nonce: int | None = None,
    ) -> InvokeV3:
        return await self._prepare_invoke_v3(
            calls,
            resource_bounds=resource_bounds,
            auto_estimate=auto_estimate,
            nonce=nonce,
        )

    async def send_transaction(
        self,
        prepared_invoke: InvokeV3,
        signature: list[int],
    ) -> SentTransactionResponse:
        signed_invoke = self._add_signature(prepared_invoke, signature)
        return await self.client.send_transaction(signed_invoke)

    async def invoke(
        self,
        contract: Contract,
        prepared_invoke: InvokeV3,
        signature: list[int],
    ) -> InvokeResult:
        invoke_transaction = self._add_signature(prepared_invoke, signature)
        res = await self.client.send_transaction(invoke_transaction)

        invoke_result = InvokeResult(
            hash=res.transaction_hash,
            _client=self.client,
            contract=contract.data,
            invoke_transaction=invoke_transaction,
        )
        return invoke_result

    async def load_contract(self, address: AddressRepresentation, is_cairo0_contract: bool = False) -> Contract:
        try:
            proxy_config = get_proxy_config() if is_cairo0_contract else False
            contract = await Contract.from_address(address=address, provider=self, proxy_config=proxy_config)
        except Exception as e:
            logging.exception(f"Error loading contract at address {hex(int(address))}: {e}")
            raise
        else:
            return contract

    async def check_multisig_required(self, contract: Contract) -> bool:
        try:
            get_signer_call = await contract.functions["getSigner"].call()
            current_signer = hex(get_signer_call.signer)  # type: ignore[union-attr]
            logging.info(f"Current signer: {current_signer}")

            get_guardian_call = await contract.functions["getGuardian"].call()
            current_guardian = hex(get_guardian_call.guardian)  # type: ignore[union-attr]
            logging.info(f"Current guardian: {current_guardian}")

            get_guardian_backup_call = await contract.functions["getGuardianBackup"].call()
            current_guardian_backup = hex(get_guardian_backup_call.guardianBackup)  # type: ignore[union-attr]
            logging.info(f"Current guardian backup: {current_guardian_backup}")

            need_multisig = current_guardian != "0x0" or current_guardian_backup != "0x0"
        except Exception as e:
            logging.exception(f"Error checking multisig requirement: {e}")
            raise
        else:
            return need_multisig

    async def process_invoke(
        self,
        contract: Contract,
        need_multisig: bool,
        prepared_invoke: InvokeV3,
        func_name: str,
    ):
        try:
            if not need_multisig:
                # Sign and send the transaction
                owner_signature = self.signer.sign_transaction(prepared_invoke)
                invoke_result = await self.invoke(contract, prepared_invoke, owner_signature)
                logging.info(f"Transaction sent with hash: {hex(invoke_result.hash)}")
                await invoke_result.wait_for_acceptance()
                logging.info("Transaction accepted on chain.")
            else:
                # Save transaction data for multisig signing
                multisig_filename = f"{func_name}_multisig.json"
                with open(multisig_filename, "w"):
                    self.print_invoke(prepared_invoke)
                logging.warning("Action requires multiple signatures.")
                logging.info(f"Prepared invoke saved to {multisig_filename}")
                logging.info(
                    "Please sign the transaction with the sign-invoke-tx command and submit with the submit-invoke-tx"
                    " command."
                )
        except Exception as e:
            logging.exception(f"Error processing invoke: {e}")
            raise

    def print_invoke(self, invoke: InvokeV3):
        invoke_schema = marshmallow_dataclass.class_schema(InvokeV3)()
        print("\n---")
        print(invoke_schema.dumps(invoke))
        print("---\n")

    def sign_message(self, typed_data: TypedData | TypedDataDict) -> list[int]:
        msg_hash = typed_data_to_message_hash(typed_data, self.address)
        r, s = message_signature(msg_hash=msg_hash, priv_key=self.signer.key_pair.private_key)  # type: ignore[attr-defined]
        return [r, s]


class StarkwareETHProxyCheck(ProxyCheck):
    async def implementation_address(self, address: Address, client: Client) -> int | None:
        return await self.get_implementation(
            address=address,
            client=client,
            get_class_func=client.get_class_hash_at,
            regex_err_msg=r"(is not deployed)",
        )

    async def implementation_hash(self, address: Address, client: Client) -> int | None:
        return await self.get_implementation(
            address=address,
            client=client,
            get_class_func=client.get_class_by_hash,
            regex_err_msg=r"(is not declared)",
        )

    @staticmethod
    async def get_implementation(
        address: Address, client: Client, get_class_func: Callable, regex_err_msg: str
    ) -> int | None:
        call = StarkwareETHProxyCheck._get_implementation_call(address=address)
        err_msg = r"(Entry point 0x[0-9a-f]+ not found in contract)|" + regex_err_msg
        try:
            (implementation,) = await client.call_contract(call=call)
            await get_class_func(implementation)
        except ClientError as err:
            if re.search(err_msg, err.message, re.IGNORECASE) or err.code == RPC_CONTRACT_ERROR:
                return None
            raise err
        return implementation

    @staticmethod
    def _get_implementation_call(address: Address) -> Call:
        return Call(
            to_addr=address,
            selector=get_selector_from_name("implementation"),
            calldata=[],
        )


def get_proxy_config():
    return ProxyConfig(
        proxy_checks=[StarkwareETHProxyCheck(), ArgentProxyCheck(), OpenZeppelinProxyCheck()],
    )



#!/usr/bin/env python3
"""
Example script demonstrating L2-only REST API usage with ParadexSubkey.

This example shows how to use REST API functionality with only L2 credentials (subkey mode),
without requiring L1 Ethereum address or private key.

Requirements:
- L2_PRIVATE_KEY: Starknet private key for the subkey
- L2_ADDRESS: L2 address of the main account (not the subkey address)

Usage:
    export L2_PRIVATE_KEY="0x..."
    export L2_ADDRESS="0x..."
    python examples/subkey_rest_api.py
"""

import os
import time
from datetime import datetime
from decimal import Decimal

from paradex_py import Paradex, ParadexSubkey
from paradex_py.common.order import Order, OrderSide, OrderType
from paradex_py.environment import TESTNET

# Environment variables
TEST_L2_ADDRESS = os.getenv("L2_ADDRESS", "")
TEST_L2_PRIVATE_KEY = os.getenv("L2_PRIVATE_KEY", "")
LOG_FILE = os.getenv("LOG_FILE", "FALSE").lower() == "true"

if LOG_FILE:
    from paradex_py.common.file_logging import file_logger

    logger = file_logger
    logger.info("Using file logger")
else:
    from paradex_py.common.console_logging import console_logger

    logger = console_logger
    logger.info("Using console logger")

# Test Public API calls
public_paradex = Paradex(env=TESTNET, logger=logger)
system_state = public_paradex.api_client.fetch_system_state()
logger.info(f"{system_state=}")
system_time = public_paradex.api_client.fetch_system_time()
logger.info(f"{system_time=}")
insurance_fund = public_paradex.api_client.fetch_insurance_fund()
logger.info(f"{insurance_fund=}")
markets = public_paradex.api_client.fetch_markets()
logger.info(f"{markets=}")
for market in markets["results"][:5]:  # Limit to 5 markets for testing
    if not int(market.get("position_limit")):
        continue
    symbol = market["symbol"]
    mkt_summary = public_paradex.api_client.fetch_markets_summary({"market": symbol})
    logger.info(f"{mkt_summary=}")
    ob = public_paradex.api_client.fetch_orderbook(market=symbol, params={"depth": 5})
    logger.info(f"{ob}=")
    bbo = public_paradex.api_client.fetch_bbo(market=symbol)
    logger.info(f"{bbo=}")

    try:
        trades = public_paradex.api_client.fetch_trades({"market": symbol, "page_size": 20})
        logger.info(f"{trades=}")
    except Exception as e:
        logger.warning(f"Could not fetch trades for {symbol}: {e}")

    try:
        funding_data = public_paradex.api_client.fetch_funding_data(params={"market": symbol})
        logger.info(f"Funding data {funding_data=}")
    except Exception as e:
        logger.warning(f"Could not fetch funding data for {symbol}: {e}")

# Test Insurance endpoints
insurance_fund = public_paradex.api_client.fetch_insurance_fund()
logger.info(f"{insurance_fund=}")

# Test System endpoints
system_config = public_paradex.api_client.fetch_system_config()
logger.info(f"{system_config=}")

# Test Private API calls with L2-only authentication using ParadexSubkey
paradex = ParadexSubkey(
    env=TESTNET,
    l2_private_key=TEST_L2_PRIVATE_KEY,
    l2_address=TEST_L2_ADDRESS,
    logger=logger,
)

account_summary = paradex.api_client.fetch_account_summary()
logger.info(f"{account_summary=}")
account_profile = paradex.api_client.fetch_account_profile()
logger.info(f"{account_profile=}")

try:
    balances = paradex.api_client.fetch_balances()
    logger.info(f"{balances=}")
except Exception as e:
    logger.warning(f"Could not fetch balances: {e}")

try:
    positions = paradex.api_client.fetch_positions()
    logger.info(f"{positions=}")
except Exception as e:
    logger.warning(f"Could not fetch positions: {e}")

try:
    transactions = paradex.api_client.fetch_transactions(params={"page_size": 5})
    logger.info(f"{transactions=}")
except Exception as e:
    logger.warning(f"Could not fetch transactions: {e}")

try:
    fills = paradex.api_client.fetch_fills(params={"page_size": 5})
    logger.info(f"{fills=}")
except Exception as e:
    logger.warning(f"Could not fetch fills: {e}")

try:
    tradebusts = paradex.api_client.fetch_tradebusts()
    logger.info(f"{tradebusts=}")
except Exception as e:
    logger.warning(f"Could not fetch tradebusts: {e}")

try:
    hist_orders = paradex.api_client.fetch_orders_history(params={"page_size": 5})
    logger.info(f"{hist_orders=}")
except Exception as e:
    logger.warning(f"Could not fetch orders history: {e}")

try:
    subaccounts = paradex.api_client.fetch_subaccounts()
    logger.info(f"{subaccounts=}")
except Exception as e:
    logger.warning(f"Could not fetch subaccounts: {e}")

try:
    account_info = paradex.api_client.fetch_account_info()
    logger.info(f"{account_info=}")
except Exception as e:
    logger.warning(f"Could not fetch account info: {e}")

points_program = paradex.api_client.fetch_points_data(
    market="ETH-USD-PERP",
    program="Maker",
)
logger.info(f"Maker {points_program=}")

transfers = paradex.api_client.fetch_transfers(params={"page_size": 5})
logger.info(f"{transfers=}")

# Per market
for market in markets["results"][:5]:  # Limit to 5 markets for testing
    if not int(market.get("position_limit")):
        continue
    symbol = market["symbol"]
    try:
        orders = paradex.api_client.fetch_orders(params={"market": symbol})
        logger.info(f"{symbol=} {orders=}")
    except Exception as e:
        logger.warning(f"Could not fetch orders for {symbol}: {e}")

    try:
        fills = paradex.api_client.fetch_fills(params={"market": symbol, "page_size": 5})
        logger.info(f"{symbol=} {fills=}")
    except Exception as e:
        logger.warning(f"Could not fetch fills for {symbol}: {e}")

    try:
        funding_payments = paradex.api_client.fetch_funding_payments(params={"market": symbol})
        logger.info(f"{symbol=} {funding_payments=}")
    except Exception as e:
        logger.warning(f"Could not fetch funding payments for {symbol}: {e}")

# Create Order object and submit order
buy_client_id = f"test_buy_{datetime.now().strftime('%Y%m%d%H%M%S')}"
buy_order = Order(
    market="BTC-USD-PERP",
    order_type=OrderType.Limit,
    order_side=OrderSide.Buy,
    size=Decimal("0.01"),
    limit_price=Decimal(11_500),
    client_id=buy_client_id,
    instruction="POST_ONLY",
    reduce_only=False,
)
response = paradex.api_client.submit_order(order=buy_order)
buy_id = response.get("id")
logger.info(f"Buy Order {response=}")
if buy_id is None:
    logger.error("Failed to get buy order ID")
    exit(1)
buy_order_status = paradex.api_client.fetch_order_by_client_id(client_id=buy_client_id)
logger.info(f"{buy_order_status=}")

# Sell order
sell_client_id = f"test_sell_{datetime.now().strftime('%Y%m%d%H%M%S')}"
sell_order = Order(
    market="ETH-USD-PERP",
    order_type=OrderType.Limit,
    order_side=OrderSide.Sell,
    size=Decimal("0.1"),
    limit_price=Decimal(5_500),
    client_id=sell_client_id,
    instruction="POST_ONLY",
    reduce_only=False,
)
response = paradex.api_client.submit_order(order=sell_order)
logger.info(f"Sell Order {response=}")
sell_id = response.get("id")
if sell_id is None:
    logger.error("Failed to get sell order ID")
    exit(1)
sell_order_status = paradex.api_client.fetch_order(order_id=sell_id)
logger.info(f"{sell_order_status=}")

# Check all open orders
orders = paradex.api_client.fetch_orders()
logger.info(f"ALL {orders=}")
logger.info("Sleeping for 3 seconds")
time.sleep(3)

# Test modify
modify_order = Order(
    order_id=buy_id,
    market="BTC-USD-PERP",
    order_type=OrderType.Limit,
    order_side=OrderSide.Buy,
    size=Decimal("0.01"),
    limit_price=Decimal(9500),
    client_id=buy_client_id,
    instruction="POST_ONLY",
    reduce_only=False,
)
response = paradex.api_client.modify_order(buy_id, modify_order)
logger.info(f"Modify order response {response}")

# Cancel ETH open order
paradex.api_client.cancel_all_orders({"market": "ETH-USD-PERP"})
orders = paradex.api_client.fetch_orders()
logger.info(f"After ETH-USD-PERP Cancel {orders=}")
paradex.api_client.cancel_order(order_id=buy_id)
orders = paradex.api_client.fetch_orders()
logger.info(f"After BUY Cancel {orders=}")

# Place Parent Order with Attached Take Profit and Stop Loss Orders
# using Batch Orders API
order = Order(
    market="BTC-USD-PERP",
    order_type=OrderType.Limit,
    order_side=OrderSide.Buy,
    size=Decimal("0.01"),
    limit_price=Decimal(95_000),
)
taker_profit_order = Order(
    market="BTC-USD-PERP",
    order_type=OrderType.TakeProfitMarket,
    order_side=OrderSide.Sell,
    size=Decimal("0.01"),
    trigger_price=Decimal(100_0000),
    reduce_only=True,
)
stop_loss_order = Order(
    market="BTC-USD-PERP",
    order_type=OrderType.StopLossMarket,
    order_side=OrderSide.Sell,
    size=Decimal("0.01"),
    trigger_price=Decimal(90_000),
    reduce_only=True,
)
orders = [
    order,
    taker_profit_order,
    stop_loss_order,
]
response = paradex.api_client.submit_orders_batch(orders=orders)
logger.info(f"Batch of orders {response=}")



from starknet_py.common import int_from_hex

from paradex_py.account.account import ParadexAccount
from paradex_py.account.utils import typed_data_to_message_hash, unflatten_signature, verify_message_signature
from paradex_py.message.auth import build_auth_message
from paradex_py.message.onboarding import build_onboarding_message
from tests.mocks.api_client import MockApiClient

TEST_L1_ADDRESS = "0xd2c7314539dCe7752c8120af4eC2AA750Cf2035e"
TEST_L1_PRIVATE_KEY = "0xf8e4d1d772cdd44e5e77615ad11cc071c94e4c06dc21150d903f28e6aa6abdff"
TEST_L2_ADDRESS = int_from_hex("0x129c135ed63df9353885e292be4426b8ed6122b13c6c0e1bb787288a1f5adfa")
TEST_L2_PRIVATE_KEY = "0x543b6cf6c91817a87174aaea4fb370ac1c694e864d7740d728f8344d53e815"
TEST_L2_PUBLIC_KEY = int_from_hex("0x2c144d2f2d4fc61b6f8967f3ba0012a87d90140bcfe5a3e92e8df83258c960f")


def test_account_l1_private_key():
    api_client = MockApiClient()
    config = api_client.fetch_system_config()

    account = ParadexAccount(
        config=config,
        l1_address=TEST_L1_ADDRESS,
        l1_private_key=TEST_L1_PRIVATE_KEY,
    )

    assert account.l2_address == TEST_L2_ADDRESS
    assert account.starknet.address == TEST_L2_ADDRESS

    assert account.l2_private_key == int_from_hex(TEST_L2_PRIVATE_KEY)
    assert account.l2_public_key == TEST_L2_PUBLIC_KEY


def test_account_l2_private_key():
    api_client = MockApiClient()
    config = api_client.fetch_system_config()

    account = ParadexAccount(
        config=config,
        l1_address=TEST_L1_ADDRESS,
        l2_private_key=TEST_L2_PRIVATE_KEY,
    )

    assert account.l2_address == TEST_L2_ADDRESS
    assert account.starknet.address == TEST_L2_ADDRESS

    assert account.l2_private_key == int_from_hex(TEST_L2_PRIVATE_KEY)
    assert account.l2_public_key == TEST_L2_PUBLIC_KEY


def test_account_onboarding_signature():
    api_client = MockApiClient()
    config = api_client.fetch_system_config()

    account = ParadexAccount(
        config=config,
        l1_address=TEST_L1_ADDRESS,
        l2_private_key=TEST_L2_PRIVATE_KEY,
    )

    sig = account.onboarding_signature()

    message = build_onboarding_message(account.l2_chain_id)
    is_signature_valid = verify_message_signature(
        typed_data_to_message_hash(message, account.l2_address),
        unflatten_signature(sig),
        account.l2_public_key,
    )
    assert is_signature_valid is True


def test_account_auth_signature():
    api_client = MockApiClient()
    config = api_client.fetch_system_config()

    account = ParadexAccount(
        config=config,
        l1_address=TEST_L1_ADDRESS,
        l2_private_key=TEST_L2_PRIVATE_KEY,
    )

    timestamp = 1706868900
    expiry = 1706955300
    sig = account.auth_signature(timestamp, expiry)

    message = build_auth_message(account.l2_chain_id, timestamp, expiry)
    is_signature_valid = verify_message_signature(
        typed_data_to_message_hash(message, account.l2_address),
        unflatten_signature(sig),
        account.l2_public_key,
    )
    assert is_signature_valid is True



import asyncio
import contextlib
import json
import logging
import time
import traceback
from collections.abc import Callable
from enum import Enum
from typing import Any, Protocol

import jwt
import websockets
from pydantic import BaseModel
from websockets import ClientConnection, State

from paradex_py.account.account import ParadexAccount
from paradex_py.constants import WS_TIMEOUT
from paradex_py.environment import Environment
from paradex_py.user_agent import get_user_agent

# Optional typed message models
try:
    from paradex_py.api.ws_message_models import validate_ws_payload
    from paradex_py.api.ws_models import validate_ws_message

    TYPED_MODELS_AVAILABLE = True
except ImportError:
    TYPED_MODELS_AVAILABLE = False

    def validate_ws_message(message_data: dict[str, Any]) -> BaseModel | None:
        return None

    def validate_ws_payload(channel_name: str, payload: dict[str, Any]) -> BaseModel | None:
        return None


class WebSocketConnection(Protocol):
    """Protocol for WebSocket-like connections."""

    async def send(self, data: str) -> None:
        """Send data to the connection."""
        ...

    async def recv(self) -> str:
        """Receive data from the connection."""
        ...

    async def close(self) -> None:
        """Close the connection."""
        ...

    @property
    def state(self) -> Any:
        """Connection state."""
        ...


class WebSocketConnector(Protocol):
    """Protocol for WebSocket connector factories."""

    async def __call__(self, url: str, headers: dict[str, str]) -> WebSocketConnection:
        """Create a WebSocket connection."""
        ...


class ParadexWebsocketChannel(Enum):
    """Enum class to define the channels for Paradex Websocket API.

    Attributes:
        ACCOUNT (str): Private websocket channel for receiving updates of account status
        BALANCE_EVENTS (str): Private websocket channel to receive PnL calculation data
        BBO (str): Public websocket channel for tick updates of orderbook best bid/ask prices and amounts
        BLOCK_TRADES (str): Private websocket channel to receive block trade updates
        FILLS (str): Private websocket channel to receive details of fills for specific account
        FUNDING_DATA (str): Public websocket channel to receive funding data updates
        FUNDING_PAYMENTS (str): Private websocket channel to receive funding payments of an account
        MARKETS_SUMMARY (str): Public websocket channel for updates of available markets
        ORDERS (str): Private websocket channel to receive order updates
        ORDER_BOOK (str): Public websocket channel for orderbook snapshot updates at most every 50ms or 100ms, optionally grouped by price tick (production only)
        POSITIONS (str): Private websocket channel to receive updates when position is changed
        TRADES (str): Public websocket channel to receive updates on trades in particular market
        TRADEBUSTS (str): Private websocket channel to receive fills that are busted by a blockchain
        TRANSACTIONS (str): Private websocket channel for receiving transaction details of fills
        TRANSFERS (str): Websocket channel for receiving transfer updates
    """

    ACCOUNT = "account"
    BALANCE_EVENTS = "balance_events"
    BBO = "bbo.{market}"
    BLOCK_TRADES = "block_trades"
    FILLS = "fills.{market}"
    FUNDING_DATA = "funding_data.{market}"
    FUNDING_PAYMENTS = "funding_payments.{market}"
    MARKETS_SUMMARY = "markets_summary"
    ORDERS = "orders.{market}"
    ORDER_BOOK = "order_book.{market}.{feed_type}@15@{refresh_rate}@{price_tick}"
    POSITIONS = "positions"
    TRADES = "trades.{market}"
    TRADEBUSTS = "tradebusts"
    TRANSACTIONS = "transaction"
    TRANSFERS = "transfers"


def _paradex_channel_prefix(value: str) -> str:
    return value.split(".")[0]


def _get_ws_channel_from_name(message_channel: str) -> ParadexWebsocketChannel | None:
    for channel in ParadexWebsocketChannel:
        if message_channel.startswith(_paradex_channel_prefix(channel.value)):
            return channel
    return None


class ParadexWebsocketClient:
    """Class to interact with Paradex WebSocket JSON-RPC API.
        Initialized along with `Paradex` class.

    Args:
        env (Environment): Environment
        logger (Optional[logging.Logger], optional): Logger. Defaults to None.
        ws_timeout (Optional[int], optional): WebSocket read timeout in seconds. Defaults to 20s.
        auto_start_reader (bool, optional): Whether to automatically start the message reader. Defaults to True.
        connector (Optional[WebSocketConnector], optional): Custom WebSocket connector for injection. Defaults to None.
        ws_url_override (Optional[str], optional): Custom WebSocket URL override. Defaults to None.
        reader_sleep_on_error (float, optional): Sleep duration after connection errors. Set to 0 for no sleep. Defaults to 1.0.
        reader_sleep_on_no_connection (float, optional): Sleep duration when no connection. Set to 0 for no sleep. Defaults to 1.0.
        validate_messages (bool, optional): Enable pydantic message validation. Requires pydantic. Defaults to False.
        ping_interval (float, optional): WebSocket ping interval in seconds. None uses websockets default. Defaults to None.
        disable_reconnect (bool, optional): Disable automatic reconnection for tight simulation control. Defaults to False.
        enable_compression (bool, optional): Enable WebSocket per-message compression (RFC 7692). Defaults to True.
        api_client (Optional[Any], optional): Reference to ParadexApiClient for token refresh. Defaults to None.

    Examples:
        >>> from paradex_py import Paradex
        >>> from paradex_py.environment import Environment
        >>> paradex = Paradex(env=Environment.TESTNET)
        >>> paradex.ws_client.connect()
        >>> # With custom timeout
        >>> from paradex_py.api.ws_client import ParadexWebsocketClient
        >>> ws_client = ParadexWebsocketClient(env=Environment.TESTNET, ws_timeout=30)
        >>> # With manual pumping disabled
        >>> ws_client = ParadexWebsocketClient(env=Environment.TESTNET, auto_start_reader=False)
        >>> # High-frequency simulator mode (no sleeps)
        >>> ws_client = ParadexWebsocketClient(env=Environment.TESTNET,
        ...                                   reader_sleep_on_error=0, reader_sleep_on_no_connection=0)
        >>> # With typed message validation
        >>> ws_client = ParadexWebsocketClient(env=Environment.TESTNET, validate_messages=True)
        >>> # With API client for automatic token refresh on reconnect
        >>> from paradex_py import Paradex
        >>> paradex = Paradex(env=Environment.TESTNET, l1_address="0x...", l1_private_key="0x...")
        >>> # WebSocket client automatically refreshes expired tokens when reconnecting
    """

    classname: str = "ParadexWebsocketClient"

    def __init__(
        self,
        env: Environment,
        logger: logging.Logger | None = None,
        ws_timeout: int | None = None,
        auto_start_reader: bool = True,
        connector: WebSocketConnector | None = None,
        ws_url_override: str | None = None,
        reader_sleep_on_error: float = 1.0,
        reader_sleep_on_no_connection: float = 1.0,
        validate_messages: bool = False,
        ping_interval: float | None = None,
        disable_reconnect: bool = False,
        enable_compression: bool = True,
        api_client: Any | None = None,
    ):
        self.env = env
        self.api_url = ws_url_override or f"wss://ws.api.{self.env}.paradex.trade/v1"
        self.logger = logger or logging.getLogger(__name__)
        self.ws: WebSocketConnection | ClientConnection | None = None
        self.account: ParadexAccount | None = None
        self.callbacks: dict[str, Callable] = {}
        self.subscribed_channels: dict[str, bool] = {}
        self.ws_timeout = ws_timeout if ws_timeout is not None else WS_TIMEOUT
        self.connector = connector
        self.auto_start_reader = auto_start_reader
        self._reader_task: asyncio.Task | None = None
        self._pending_reconnect_task: asyncio.Task | None = None  # Track pending reconnect tasks
        self._is_closing: bool = False  # Flag to prevent reconnection during intentional closure

        # Lock to synchronize WebSocket recv() calls between background reader and manual pump_once
        self._recv_lock = asyncio.Lock()

        # Configurable sleep durations for simulator-friendly behavior
        self.reader_sleep_on_error = reader_sleep_on_error
        self.reader_sleep_on_no_connection = reader_sleep_on_no_connection

        # Heartbeat and reconnection control
        self.ping_interval = ping_interval
        self.disable_reconnect = disable_reconnect

        # Compression control
        self.enable_compression = enable_compression

        # Optional message validation
        self.validate_messages = validate_messages and TYPED_MODELS_AVAILABLE

        # API client reference for token refresh on reconnect
        self._api_client = api_client

        if auto_start_reader:
            try:
                loop = asyncio.get_event_loop()
                if loop.is_running():
                    self._reader_task = loop.create_task(self._read_messages())
                else:
                    # Event loop exists but not running, don't start task yet
                    self._reader_task = None
            except RuntimeError:
                # No event loop running, will start reader when connect() is called
                self._reader_task = None

    async def __aexit__(self):
        await self._close_connection()

    def init_account(self, account: ParadexAccount) -> None:
        self.account = account

    def _build_connect_kwargs(self, extra_headers: dict[str, str]) -> dict[str, Any]:
        """Build connection kwargs for websockets.connect()."""
        connect_kwargs: dict[str, Any] = {
            "additional_headers": extra_headers,
        }
        if self.ping_interval is not None:
            connect_kwargs["ping_interval"] = int(self.ping_interval)
        # Configure compression
        if not self.enable_compression:
            connect_kwargs["compression"] = None
        return connect_kwargs

    async def connect(self) -> bool:
        """Connect to Paradex WebSocket API.

        Returns:
            bool: True if connection is successful.

        Examples:
            >>> from paradex_py import Paradex
            >>> from paradex_py.environment import TESTNET
            >>> async def main():
            ...     paradex = Paradex(env=TESTNET)
            ...     await paradex.ws_client.connect()
            >>> import asyncio
            >>> asyncio.run(main())
        """

        try:
            self.subscribed_channels = {}
            extra_headers = {"User-Agent": get_user_agent()}
            if self.account:
                extra_headers.update({"Authorization": f"Bearer {self.account.jwt_token}"})

            # Use custom connector if provided, otherwise use default websockets.connect
            if self.connector is not None:
                self.ws = await self.connector(self.api_url, extra_headers)
            else:
                connect_kwargs = self._build_connect_kwargs(extra_headers)
                self.ws = await websockets.connect(self.api_url, **connect_kwargs)

            self.logger.info(f"{self.classname}: Connected to {self.api_url}")

            # Start reader task if auto_start_reader is enabled and not already running
            if self.auto_start_reader and self._reader_task is None:
                self._reader_task = asyncio.create_task(self._read_messages())

            if self.account:
                await self._send_auth_id(self.ws, self.account.jwt_token)
                self.logger.info(f"{self.classname}: Authenticated to {self.api_url}")
        except (
            websockets.exceptions.ConnectionClosedOK,
            websockets.exceptions.ConnectionClosed,
        ):
            self.logger.exception(f"{self.classname}: Connection already closed")
            self.ws = None
        except Exception:
            self.logger.exception(f"{self.classname}: traceback:{traceback.format_exc()}")
            self.ws = None

        # Check connection state - handle both websockets.State and custom connection states
        is_connected = False
        if self.ws is not None:
            is_connected = self.ws.state == State.OPEN if hasattr(self.ws.state, "value") else hasattr(self.ws, "state")

        return is_connected

    async def close(self):
        """Close the WebSocket connection and clean up resources.

        This method should be called when done using the WebSocket client
        to properly clean up connections and background tasks.

        Examples:
            >>> import asyncio
            >>> from paradex_py.api.ws_client import ParadexWebsocketClient
            >>> from paradex_py.environment import Environment
            >>> async def main():
            ...     ws_client = ParadexWebsocketClient(env=Environment.TESTNET)
            ...     try:
            ...         await ws_client.connect()
            ...         # Use websocket client
            ...     finally:
            ...         await ws_client.close()
            >>> asyncio.run(main())
        """
        await self._close_connection()

    async def _close_connection(self):
        try:
            # Set flag to prevent reconnection during intentional closure
            self._is_closing = True

            # Cancel reader task if it exists
            if self._reader_task and not self._reader_task.done():
                self._reader_task.cancel()
                with contextlib.suppress(asyncio.CancelledError):
                    await self._reader_task
                self._reader_task = None

            if self.ws:
                self.logger.info(f"{self.classname}: Closing connection...")
                await self.ws.close()
                self.logger.info(f"{self.classname}: Connection closed")
            else:
                self.logger.info(f"{self.classname}: No connection to close")
        except Exception:
            self.logger.exception(f"{self.classname}: Error thrown when closing connection {traceback.format_exc()}")
        finally:
            # Reset flag after closing is complete
            self._is_closing = False

    def _decode_jwt_payload(self, token: str) -> dict[str, Any] | None:
        """Decode JWT token payload without signature verification.

        Args:
            token: JWT token string

        Returns:
            Decoded payload dict, or None if decoding fails
        """
        try:
            return jwt.decode(token, options={"verify_signature": False})
        except jwt.exceptions.DecodeError as e:
            self.logger.warning(f"{self.classname}: Failed to decode JWT token: {e}")
            return None

    def _is_token_expired(self) -> bool:
        """Check if JWT token has expired by decoding the token and checking exp claim.

        Returns True if token has expired or will expire soon (within 60 seconds safety margin).
        Returns False if we can't determine expiration.
        """
        if not self.account or not self.account.jwt_token:
            # No token available
            return False

        payload = self._decode_jwt_payload(self.account.jwt_token)
        if not payload or "exp" not in payload:
            # Can't determine expiration from token
            return False

        # Check if token has expired or will expire within 60 seconds (safety margin)
        exp_time = payload["exp"]
        current_time = time.time()
        return current_time >= (exp_time - 60)

    async def _reconnect(self):
        if self.disable_reconnect:
            self.logger.info(f"{self.classname}: Reconnection disabled, skipping...")
            return

        try:
            self.logger.info(f"{self.classname}: Reconnect websocket...")

            # Check if token has expired before reconnecting
            if self._is_token_expired() and self._api_client and hasattr(self._api_client, "auth"):
                try:
                    self.logger.info(f"{self.classname}: Token expired, refreshing before reconnection")
                    self._api_client.auth()
                    self.logger.info(f"{self.classname}: Token refreshed successfully")
                except Exception:
                    self.logger.exception(f"{self.classname}: Failed to refresh token: {traceback.format_exc()}")
                    # Continue with reconnection anyway - server will reject if token is invalid

            # Acquire lock to ensure no concurrent recv() operations during reconnection
            async with self._recv_lock:
                await self._close_connection()
                await self.connect()
            await self._resubscribe()
        except Exception:
            self.logger.exception(f"{self.classname}: Reconnect failed {traceback.format_exc()}")

    async def _reconnect_with_auth_refresh(self):
        """Reconnect with fresh authentication token.

        This method is called when an authentication error is detected (e.g., expired JWT).
        It forces a token refresh before reconnecting.
        """
        if self.disable_reconnect:
            self.logger.info(f"{self.classname}: Reconnection disabled, skipping auth refresh reconnect...")
            return

        try:
            self.logger.info(f"{self.classname}: Reconnecting due to auth error...")

            # Force token refresh via API client if available
            if self._api_client and hasattr(self._api_client, "auth"):
                try:
                    self.logger.info(f"{self.classname}: Forcing JWT token refresh due to auth error")
                    self._api_client.auth()
                    self.logger.info(f"{self.classname}: Token refreshed successfully")
                except Exception:
                    self.logger.exception(f"{self.classname}: Failed to refresh token: {traceback.format_exc()}")
                    # Continue with reconnection anyway - server will reject if token is invalid

            # Now reconnect with the refreshed token
            await self._reconnect()
        except Exception:
            self.logger.exception(f"{self.classname}: Auth refresh reconnect failed {traceback.format_exc()}")

    async def _resubscribe(self):
        if self.ws and self.ws.state == State.OPEN:
            for channel_name in self.callbacks:
                await self._subscribe_to_channel_by_name(channel_name)
        else:
            self.logger.warning(f"{self.classname}: Resubscribe - No connection")

    async def _send_auth_id(
        self,
        websocket: WebSocketConnection | ClientConnection,
        paradex_jwt: str,
    ) -> None:
        """
        Sends an authentication message to the Paradex WebSocket.
        """
        await websocket.send(
            json.dumps(
                {
                    "id": int(time.time() * 1_000_000),
                    "jsonrpc": "2.0",
                    "method": "auth",
                    "params": {"bearer": paradex_jwt},
                }
            )
        )

    def _check_subscribed_channel(self, message: dict) -> None:
        if "id" in message:
            # Check for successful subscription
            channel_subscribed: str | None = message.get("result", {}).get("channel")
            if channel_subscribed:
                self.logger.info(f"{self.classname}: Subscribed to channel:{channel_subscribed}")
                self.subscribed_channels[channel_subscribed] = True
            # Check for subscription error
            error_info = message.get("error")
            if error_info:
                error_code = error_info.get("code", "unknown")
                error_message = error_info.get("message", "unknown error")
                self.logger.error(f"{self.classname}: Subscription failed - code:{error_code} message:{error_message}")

                # Check for auth-related errors (invalid bearer token)
                if error_code == 40111 or "bearer token" in str(error_message).lower():
                    self.logger.error(f"{self.classname}: Authentication error detected - JWT token may be expired")
                    # Trigger reconnection with fresh token
                    if not self.disable_reconnect:
                        self._pending_reconnect_task = asyncio.create_task(self._reconnect_with_auth_refresh())

                # Note: We don't mark the channel as subscribed since it failed

    def _is_connection_open(self) -> bool:
        """Check if WebSocket connection is open - handle both websockets and custom connections."""
        if not self.ws:
            return False

        if hasattr(self.ws.state, "value"):
            # websockets.State enum
            return self.ws.state == State.OPEN
        else:
            # Custom connection - check if state indicates open
            state_val = getattr(self.ws.state, "value", None) if hasattr(self.ws, "state") else None
            return state_val == "OPEN" or (state_val is None and hasattr(self.ws, "recv"))

    async def _receive_and_process_message(self) -> None:
        """Receive and process a single WebSocket message."""
        if self.ws is None:
            raise RuntimeError("WebSocket not connected")
        async with self._recv_lock:
            response = await asyncio.wait_for(self.ws.recv(), timeout=self.ws_timeout)
        if isinstance(response, bytes):
            response = response.decode("utf-8")
        await self._process_message(response)

    async def _handle_message_receive_error(self, error: Exception) -> None:
        """Handle errors that occur while receiving messages."""
        if isinstance(error, websockets.exceptions.ConnectionClosedError | websockets.exceptions.ConnectionClosedOK):
            # Don't reconnect if we're intentionally closing the connection
            if self._is_closing:
                self.logger.info(
                    f"{self.classname}: Connection closed during intentional closure, skipping reconnection"
                )
                return
            self.logger.exception(f"{self.classname}: Connection closed traceback:{traceback.format_exc()}")
            await self._reconnect()
        elif isinstance(error, asyncio.TimeoutError):
            pass
        elif isinstance(error, asyncio.CancelledError):
            self.logger.info(f"{self.classname}: Reader task cancelled")
            raise
        else:
            self.logger.exception(f"{self.classname}: Connection failed traceback:{traceback.format_exc()}")
            if self.reader_sleep_on_error > 0:
                await asyncio.sleep(self.reader_sleep_on_error)

    async def _read_messages(self) -> None:
        try:
            while True:
                if self._is_connection_open() and self.ws is not None:
                    try:
                        await self._receive_and_process_message()
                    except Exception as e:
                        await self._handle_message_receive_error(e)
                elif self.reader_sleep_on_no_connection > 0:
                    await asyncio.sleep(self.reader_sleep_on_no_connection)
        except asyncio.CancelledError:
            # Re-raise cancellation to allow proper cleanup
            self.logger.info(f"{self.classname}: Reader task cancelled, cleaning up")
            raise
        except Exception:
            self.logger.exception(f"{self.classname}: Unexpected error in reader task: {traceback.format_exc()}")
            raise

    async def _process_message(self, response: str) -> None:
        """Process a single WebSocket message."""
        message = json.loads(response)
        self._check_subscribed_channel(message)
        if "params" not in message:
            self.logger.debug(f"{self.classname}: Non-actionable message:{message}")
        else:
            message_channel = message["params"].get("channel")
            ws_channel: ParadexWebsocketChannel | None = _get_ws_channel_from_name(message_channel)

            # Optional WebSocket RPC message validation
            if self.validate_messages:
                validated_message = validate_ws_message(message)
                if validated_message is not None:
                    # Use validated message structure
                    message = validated_message.model_dump() if hasattr(validated_message, "model_dump") else message
                    self.logger.debug(f"{self.classname}: WebSocket RPC message validated")
                else:
                    self.logger.warning(f"{self.classname}: WebSocket RPC message validation failed")

                # Validate payload against AsyncAPI models
                if "params" in message and "data" in message:
                    channel_name = message["params"].get("channel", "")
                    payload_data = message["data"]
                    validated_payload = validate_ws_payload(channel_name, payload_data)
                    if validated_payload is not None:
                        # Replace data with validated payload
                        message["data"] = validated_payload.model_dump()
                        self.logger.debug(f"{self.classname}: WebSocket payload validated for channel {channel_name}")
                    else:
                        self.logger.warning(
                            f"{self.classname}: WebSocket payload validation failed for channel {channel_name}"
                        )

            if ws_channel is None:
                self.logger.debug(f"{self.classname}: unregistered channel:{message_channel} message:{message}")
            elif message_channel in self.callbacks:
                self.logger.debug(
                    f"{self.classname}: channel:{message_channel}"
                    f" callback:{self.callbacks[message_channel]}"
                    f" message:{message}"
                )
                await self.callbacks[message_channel](ws_channel, message)
            else:
                self.logger.info(f"{self.classname}: Non-callback channel:{message_channel}")

    async def pump_once(self) -> bool:
        """Manually pump one message from the WebSocket connection.

        Returns:
            bool: True if a message was processed, False if no message available or connection closed.
        """
        if not self.ws:
            return False

        try:
            # Try to receive with a very short timeout to avoid blocking
            async with self._recv_lock:
                response = await asyncio.wait_for(self.ws.recv(), timeout=0.001)
        except asyncio.TimeoutError:
            return False
        except Exception:
            self.logger.exception(f"{self.classname}: Error in pump_once: {traceback.format_exc()}")
            return False
        else:
            if isinstance(response, bytes):
                response = response.decode("utf-8")
            await self._process_message(response)
            return True

    async def inject(self, message: str) -> None:
        """Inject a raw message string into the message processing pipeline.

        Args:
            message: Raw JSON string to process as if received from WebSocket.
        """
        try:
            await self._process_message(message)
        except Exception:
            self.logger.exception(f"{self.classname}: Error in inject: {traceback.format_exc()}")

    async def _send(self, message: str):
        try:
            if self.ws:
                await self.ws.send(message)
        except websockets.exceptions.ConnectionClosedError as e:
            self.logger.info(f"{self.classname}: Restarted connection error:{e}")
            await self._reconnect()
            if self.ws:
                await self.ws.send(message)
        except Exception:
            self.logger.exception(f"{self.classname}: Send failed traceback:{traceback.format_exc()}")
            await self._reconnect()

    async def subscribe(
        self,
        channel: ParadexWebsocketChannel,
        callback: Callable,
        params: dict | None = None,
    ) -> None:
        """Subscribe to a websocket channel with optional parameters.
            Callback function is invoked when a message is received.

        Args:
            channel (ParadexWebsocketChannel): Channel to subscribe
            callback (Callable): Callback function
            params (Optional[dict], optional): Parameters for the channel. Defaults to None.

        Examples:
        >>> from paradex_py import Paradex
        >>> from paradex_py.environment import TESTNET
        >>> from paradex_py.api.ws_client import ParadexWebsocketChannel, ParadexWebsocketClient
        >>> async def main():
        ...     async def on_message(ws_channel, message):
        ...         print(ws_channel, message)
        ...     paradex = Paradex(env=TESTNET)
        ...     await paradex.ws_client.connect()
        ...     await paradex.ws_client.subscribe(ParadexWebsocketChannel.MARKETS_SUMMARY, callback=on_message)
        >>> import asyncio
        >>> asyncio.run(main())
        """
        if params is None:
            params = {}
        # Note: Set default to all markets if no params are provided which
        # allows backward compatibility with old market_summary where
        # no params were required.
        if channel == ParadexWebsocketChannel.MARKETS_SUMMARY and not params:
            params = {"market": "ALL"}

        # Handle ORDER_BOOK channel with optional parameters
        if channel == ParadexWebsocketChannel.ORDER_BOOK:
            # Set defaults for required parameters
            format_params = params.copy()
            if "feed_type" not in format_params:
                format_params["feed_type"] = "snapshot"
            if "refresh_rate" not in format_params:
                format_params["refresh_rate"] = "100ms"
            # price_tick is optional - if not provided or empty, omit it
            if "price_tick" not in format_params or not format_params["price_tick"]:
                format_params.pop("price_tick", None)
                base_format = "order_book.{market}.{feed_type}@15@{refresh_rate}"
            else:
                base_format = "order_book.{market}.{feed_type}@15@{refresh_rate}@{price_tick}"
            channel_name = base_format.format(**format_params)
        else:
            channel_name = channel.value.format(**params)
        self.callbacks[channel_name] = callback
        self.logger.info(f"{self.classname}: Subscribe channel:{channel_name} params:{params} callback:{callback}")
        await self._subscribe_to_channel_by_name(channel_name)

    async def subscribe_by_name(
        self,
        channel_name: str,
        callback: Callable | None = None,
    ) -> None:
        """Subscribe to a channel by exact name string.

        This is useful for simulation tooling where you want to subscribe
        to channels with exact names without using the enum formatting.

        Args:
            channel_name: Exact channel name (e.g., "bbo.BTC-USD-PERP")
            callback: Optional callback function. If provided, registers the callback.
        """
        if callback is not None:
            self.callbacks[channel_name] = callback
            self.logger.info(f"{self.classname}: Subscribe by name channel:{channel_name} callback:{callback}")

        await self._subscribe_to_channel_by_name(channel_name)

    async def unsubscribe_by_name(self, channel_name: str) -> None:
        """Unsubscribe from a channel by exact name string.

        Symmetric with subscribe_by_name for complete channel lifecycle control.

        Args:
            channel_name: Exact channel name (e.g., "bbo.BTC-USD-PERP")
        """
        # Remove from subscribed channels and callbacks
        self.subscribed_channels.pop(channel_name, None)
        self.callbacks.pop(channel_name, None)

        self.logger.info(f"{self.classname}: Unsubscribe by name channel:{channel_name}")

        # Send unsubscribe message
        unsubscribe_message = {
            "jsonrpc": "2.0",
            "method": "unsubscribe",
            "params": {"channel": channel_name},
            "id": str(int(time.time() * 1_000_000)),
        }
        await self._send(json.dumps(unsubscribe_message))

    def get_subscriptions(self) -> dict[str, bool]:
        """Get current subscription map.

        Returns:
            Dictionary mapping channel names to subscription status
        """
        return self.subscribed_channels.copy()

    async def pump_until(self, predicate: Callable[[dict], bool], timeout_s: float = 10.0) -> int:
        """Deterministic consumption helper for simulators.

        Pumps messages until predicate returns True or timeout is reached.
        Useful for waiting for specific message conditions in tests.

        Args:
            predicate: Function that takes a message dict and returns True to stop
            timeout_s: Maximum time to wait in seconds

        Returns:
            Number of messages processed before predicate was satisfied or timeout

        Examples:
            # Wait for BBO message with specific price
            count = await ws_client.pump_until(
                lambda msg: msg.get('params', {}).get('channel', '').startswith('bbo')
                           and float(msg.get('data', {}).get('bid', 0)) > 50000,
                timeout_s=5.0
            )
        """
        start_time = time.time()
        message_count = 0
        last_message = None

        # Set up temporary message capture
        def capture_message(channel: str, message: dict):
            nonlocal last_message
            last_message = message

        # Store original callbacks to restore later
        original_callbacks = self.callbacks.copy()

        try:
            while time.time() - start_time < timeout_s:
                # Try to pump one message
                if await self.pump_once():
                    message_count += 1
                    # Check if predicate is satisfied with the last message
                    if last_message and predicate(last_message):
                        break
                    else:
                        # Continue processing if predicate not satisfied
                        pass
                else:
                    # No message available, small delay to avoid busy waiting
                    await asyncio.sleep(0.001)
        finally:
            # Restore original callbacks
            self.callbacks = original_callbacks

        return message_count

    async def _subscribe_to_channel_by_name(
        self,
        channel_name: str,
    ) -> None:
        await self._send(
            json.dumps(
                {
                    "id": int(time.time() * 1_000_000),
                    "jsonrpc": "2.0",
                    "method": "subscribe",
                    "params": {"channel": channel_name},
                }
            )
        )



from unittest.mock import Mock, patch

import pytest

from paradex_py.api.api_client import ParadexApiClient
from paradex_py.environment import TESTNET


class TestParadexApiClient:
    def setup_method(self):
        """Setup method to create an API client instance for each test."""
        self.api_client = ParadexApiClient(env=TESTNET)
        # Mock the account and auth methods
        self.api_client.account = Mock()
        self.api_client._validate_auth = Mock()

    def test_cancel_orders_batch_with_order_ids(self):
        """Test cancel_orders_batch with order IDs."""
        order_ids = ["order1", "order2", "order3"]
        expected_payload = {"order_ids": order_ids}

        # Mock the delete method
        with patch.object(self.api_client, "delete") as mock_delete:
            mock_delete.return_value = {"results": []}

            self.api_client.cancel_orders_batch(order_ids=order_ids)

            # Verify the delete method was called with correct parameters
            mock_delete.assert_called_once_with(
                api_url=self.api_client.api_url, path="orders/batch", params=None, payload=expected_payload
            )
            # Verify auth validation was called
            self.api_client._validate_auth.assert_called_once()

    def test_cancel_orders_batch_with_client_order_ids(self):
        """Test cancel_orders_batch with client order IDs."""
        client_order_ids = ["client1", "client2"]
        expected_payload = {"client_order_ids": client_order_ids}

        with patch.object(self.api_client, "delete") as mock_delete:
            mock_delete.return_value = {"results": []}

            self.api_client.cancel_orders_batch(client_order_ids=client_order_ids)

            mock_delete.assert_called_once_with(
                api_url=self.api_client.api_url, path="orders/batch", params=None, payload=expected_payload
            )

    def test_cancel_orders_batch_with_both_ids(self):
        """Test cancel_orders_batch with both order IDs and client order IDs."""
        order_ids = ["order1"]
        client_order_ids = ["client1"]
        expected_payload = {"order_ids": order_ids, "client_order_ids": client_order_ids}

        with patch.object(self.api_client, "delete") as mock_delete:
            mock_delete.return_value = {"results": []}

            self.api_client.cancel_orders_batch(order_ids=order_ids, client_order_ids=client_order_ids)

            mock_delete.assert_called_once_with(
                api_url=self.api_client.api_url, path="orders/batch", params=None, payload=expected_payload
            )

    def test_cancel_orders_batch_no_ids_raises_error(self):
        """Test cancel_orders_batch raises error when no IDs provided."""
        with pytest.raises(Exception) as exc_info:
            self.api_client.cancel_orders_batch()

        # The raise_value_error function should raise an exception
        assert "Must provide either order_ids or client_order_ids" in str(exc_info.value)

    def test_fetch_klines_with_required_params(self):
        """Test fetch_klines with required parameters."""
        symbol = "BTC-USD-PERP"
        resolution = "5"
        start_at = 1640995200000
        end_at = 1641081600000

        expected_params = {
            "symbol": symbol,
            "resolution": resolution,
            "start_at": start_at,
            "end_at": end_at,
        }

        with patch.object(self.api_client, "get") as mock_get:
            mock_get.return_value = {"results": []}

            self.api_client.fetch_klines(symbol=symbol, resolution=resolution, start_at=start_at, end_at=end_at)

            mock_get.assert_called_once_with(
                api_url=self.api_client.api_url, path="markets/klines", params=expected_params
            )

    def test_fetch_klines_with_price_kind(self):
        """Test fetch_klines with optional price_kind parameter."""
        symbol = "ETH-USD-PERP"
        resolution = "15"
        start_at = 1640995200000
        end_at = 1641081600000
        price_kind = "mark"

        expected_params = {
            "symbol": symbol,
            "resolution": resolution,
            "start_at": start_at,
            "end_at": end_at,
            "price_kind": price_kind,
        }

        with patch.object(self.api_client, "get") as mock_get:
            mock_get.return_value = {"results": []}

            self.api_client.fetch_klines(
                symbol=symbol, resolution=resolution, start_at=start_at, end_at=end_at, price_kind=price_kind
            )

            mock_get.assert_called_once_with(
                api_url=self.api_client.api_url, path="markets/klines", params=expected_params
            )

    def test_fetch_klines_without_price_kind(self):
        """Test fetch_klines without optional price_kind parameter."""
        symbol = "SOL-USD-PERP"
        resolution = "60"
        start_at = 1640995200000
        end_at = 1641081600000

        expected_params = {
            "symbol": symbol,
            "resolution": resolution,
            "start_at": start_at,
            "end_at": end_at,
        }

        with patch.object(self.api_client, "get") as mock_get:
            mock_get.return_value = {"results": []}

            self.api_client.fetch_klines(symbol=symbol, resolution=resolution, start_at=start_at, end_at=end_at)

            mock_get.assert_called_once_with(
                api_url=self.api_client.api_url, path="markets/klines", params=expected_params
            )

            # Verify price_kind is not in params when not provided
            called_params = mock_get.call_args[1]["params"]
            assert "price_kind" not in called_params



"""
Injection protocols and type contracts for paradex_py.

This module exports all the protocols that define injection points
for custom implementations in simulation, testing, and production environments.
"""

from typing import Any, Protocol

import httpx


# WebSocket protocols
class WebSocketConnection(Protocol):
    """Protocol for WebSocket connection implementations."""

    async def send(self, data: str) -> None:
        """Send data through the WebSocket."""
        ...

    async def recv(self) -> str:
        """Receive data from the WebSocket."""
        ...

    async def close(self) -> None:
        """Close the WebSocket connection."""
        ...

    @property
    def state(self) -> Any:
        """Get connection state."""
        ...


class WebSocketConnector(Protocol):
    """Protocol for WebSocket connector implementations."""

    async def __call__(self, url: str, headers: dict[str, str]) -> WebSocketConnection:
        """Create a WebSocket connection.

        Args:
            url: WebSocket URL to connect to
            headers: Connection headers

        Returns:
            WebSocket connection instance
        """
        ...


# HTTP protocols
class HttpClientLike(Protocol):
    """Protocol for HTTP client implementations."""

    def request(
        self,
        method: str,
        url: str,
        *,
        params: dict[str, Any] | None = None,
        json: Any = None,
        headers: dict[str, str] | None = None,
        timeout: float | None = None,
    ) -> httpx.Response:
        """Make HTTP request."""
        ...


class TransportLike(Protocol):
    """Protocol for HTTP transport implementations."""

    def handle_request(self, request: httpx.Request) -> httpx.Response:
        """Handle HTTP request and return response."""
        ...


class RetryStrategy(Protocol):
    """Protocol for retry/backoff strategies."""

    def should_retry(self, attempt: int, response: httpx.Response | None, exception: Exception | None) -> bool:
        """Determine if request should be retried.

        Args:
            attempt: Current attempt number (0-based)
            response: HTTP response if available
            exception: Exception if request failed

        Returns:
            True if request should be retried
        """
        ...

    def get_delay(self, attempt: int) -> float:
        """Get delay before next retry attempt.

        Args:
            attempt: Current attempt number (0-based)

        Returns:
            Delay in seconds
        """
        ...


class RequestHook(Protocol):
    """Protocol for request/response observability hooks."""

    def on_request(self, method: str, url: str, headers: dict[str, Any] | None) -> None:
        """Called before making request.

        Args:
            method: HTTP method
            url: Request URL (potentially redacted)
            headers: Request headers (potentially redacted)
        """
        ...

    def on_response(self, method: str, url: str, status_code: int, duration_ms: float) -> None:
        """Called after receiving response.

        Args:
            method: HTTP method
            url: Request URL (potentially redacted)
            status_code: HTTP status code
            duration_ms: Request duration in milliseconds
        """
        ...


class AuthProvider(Protocol):
    """Protocol for custom authentication flows."""

    def get_token(self) -> str | None:
        """Get current JWT token.

        Returns:
            JWT token string or None if not available
        """
        ...

    def refresh_if_needed(self) -> str | None:
        """Refresh token if needed and return current token.

        Returns:
            JWT token string or None if refresh failed
        """
        ...


# Signing protocols
class Signer(Protocol):
    """Protocol for order signing implementations."""

    def sign_order(self, order_data: dict[str, Any]) -> dict[str, Any]:
        """Sign an order.

        Args:
            order_data: Order data to sign

        Returns:
            Signed order data with signature fields
        """
        ...

    def sign_batch(self, orders: list[dict[str, Any]]) -> list[dict[str, Any]]:
        """Sign multiple orders.

        Args:
            orders: List of orders to sign

        Returns:
            List of signed orders
        """
        ...


# Default implementations
class DefaultRetryStrategy:
    """Default exponential backoff retry strategy."""

    def __init__(self, max_retries: int = 3, base_delay: float = 1.0, max_delay: float = 60.0):
        self.max_retries = max_retries
        self.base_delay = base_delay
        self.max_delay = max_delay

    def should_retry(self, attempt: int, response: Any | None, exception: Exception | None) -> bool:
        if attempt >= self.max_retries:
            return False

        # Retry on network errors
        if exception is not None:
            return True

        # Retry on 5xx errors and rate limits
        if response is not None:
            return response.status_code >= 500 or response.status_code == 429

        return False

    def get_delay(self, attempt: int) -> float:
        delay = self.base_delay * (2**attempt)
        return min(delay, self.max_delay)


class NoOpSigner:
    """No-op signer for full simulation scenarios."""

    def sign_order(self, order_data: dict[str, Any]) -> dict[str, Any]:
        """Return order data without signing."""
        return order_data

    def sign_batch(self, orders: list[dict[str, Any]]) -> list[dict[str, Any]]:
        """Return orders without signing."""
        return orders


# Type aliases for common injection types
ClientConnectionLike = WebSocketConnection
WebsocketConnectorLike = WebSocketConnector

# Export all protocols
__all__ = [
    # WebSocket protocols
    "WebSocketConnection",
    "WebSocketConnector",
    "ClientConnectionLike",  # Alias
    "WebsocketConnectorLike",  # Alias
    # HTTP protocols
    "HttpClientLike",
    "TransportLike",
    "RetryStrategy",
    "RequestHook",
    # Auth protocols
    "AuthProvider",
    # Signing protocols
    "Signer",
    "NoOpSigner",
    # Default implementations
    "DefaultRetryStrategy",
]



from paradex_py.message.onboarding import build_onboarding_message


def test_build_onboarding_message():
    assert build_onboarding_message(1) == {
        "message": {
            "action": "Onboarding",
        },
        "domain": {"name": "Paradex", "chainId": "0x1", "version": "1"},
        "primaryType": "Constant",
        "types": {
            "StarkNetDomain": [
                {"name": "name", "type": "felt"},
                {"name": "chainId", "type": "felt"},
                {"name": "version", "type": "felt"},
            ],
            "Constant": [
                {"name": "action", "type": "felt"},
            ],
        },
    }



"""Tests for HTTP compression support in HttpClient."""

import httpx

from paradex_py.api.http_client import HttpClient
from paradex_py.environment import TESTNET
from paradex_py.paradex import Paradex


class TestHttpCompression:
    """Test suite for HTTP compression configuration."""

    def test_compression_enabled_by_default(self):
        """Test that compression is enabled by default (httpx default behavior)."""
        http_client = HttpClient()

        # When compression is enabled, httpx automatically adds Accept-Encoding header
        # or doesn't explicitly set "identity"
        assert (
            "Accept-Encoding" not in http_client.client.headers
            or http_client.client.headers["Accept-Encoding"] != "identity"
        )

    def test_compression_explicitly_enabled(self):
        """Test that compression can be explicitly enabled."""
        http_client = HttpClient(enable_compression=True)

        # Should not set Accept-Encoding to identity
        assert (
            "Accept-Encoding" not in http_client.client.headers
            or http_client.client.headers["Accept-Encoding"] != "identity"
        )

    def test_compression_disabled(self):
        """Test that compression can be disabled."""
        http_client = HttpClient(enable_compression=False)

        # Should set Accept-Encoding to identity
        assert http_client.client.headers["Accept-Encoding"] == "identity"

    def test_compression_disabled_with_custom_client(self):
        """Test that compression can be disabled with custom injected client."""
        custom_client = httpx.Client()
        http_client = HttpClient(http_client=custom_client, enable_compression=False)

        # Should set Accept-Encoding to identity on custom client
        assert http_client.client.headers["Accept-Encoding"] == "identity"

    def test_compression_enabled_with_custom_client(self):
        """Test that compression is left as default with custom client when enabled."""
        custom_client = httpx.Client()
        http_client = HttpClient(http_client=custom_client, enable_compression=True)

        # Should not modify Accept-Encoding header when compression is enabled
        assert (
            "Accept-Encoding" not in http_client.client.headers
            or http_client.client.headers["Accept-Encoding"] != "identity"
        )

    def test_compression_overrides_custom_accept_encoding_when_disabled(self):
        """Test that disabling compression overrides custom Accept-Encoding header."""
        custom_client = httpx.Client()
        custom_client.headers["Accept-Encoding"] = "custom-encoding"
        http_client = HttpClient(http_client=custom_client, enable_compression=False)

        # When compression is explicitly disabled, should override custom Accept-Encoding
        assert http_client.client.headers["Accept-Encoding"] == "identity"

    def test_compression_disabled_via_paradex_constructor(self):
        """Test that compression can be disabled via Paradex constructor."""
        paradex = Paradex(env=TESTNET, enable_http_compression=False)

        # Should have created an HttpClient with compression disabled
        assert paradex.api_client.client.headers["Accept-Encoding"] == "identity"

    def test_compression_enabled_via_paradex_constructor(self):
        """Test that compression is enabled via Paradex constructor."""
        paradex = Paradex(env=TESTNET, enable_http_compression=True)

        # Should not have Accept-Encoding set to identity
        assert (
            "Accept-Encoding" not in paradex.api_client.client.headers
            or paradex.api_client.client.headers["Accept-Encoding"] != "identity"
        )

    def test_compression_default_via_paradex_constructor(self):
        """Test that compression is enabled by default when not specified."""
        paradex = Paradex(env=TESTNET)

        # Should use default compression behavior (not set to identity)
        # Note: Paradex doesn't create HttpClient when all parameters are default
        # So we check the underlying httpx client in the API client
        assert (
            "Accept-Encoding" not in paradex.api_client.client.headers
            or paradex.api_client.client.headers["Accept-Encoding"] != "identity"
        )

    def test_compression_with_other_http_options(self):
        """Test that compression works alongside other HTTP options."""
        http_client = HttpClient(enable_compression=False, default_timeout=30.0)

        # Should have both compression disabled and timeout set
        assert http_client.client.headers["Accept-Encoding"] == "identity"
        assert http_client.default_timeout == 30.0



import time


def time_now_milli_secs() -> int:
    return int(time.time() * 1_000)


def time_now_micro_secs() -> int:
    return int(time.time() * 1_000_000)


def raise_value_error(message: str):
    raise ValueError(message)



import asyncio
import os
from decimal import Decimal

from starknet_py.common import int_from_hex

from paradex_py import Paradex
from paradex_py.environment import TESTNET

# Environment variables
TEST_L1_ADDRESS = os.getenv("L1_ADDRESS", "")
TEST_L1_PRIVATE_KEY = int_from_hex(os.getenv("L1_PRIVATE_KEY", ""))
LOG_FILE = os.getenv("LOG_FILE", "FALSE").lower() == "true"

if LOG_FILE:
    from paradex_py.common.file_logging import file_logger

    logger = file_logger
else:
    from paradex_py.common.console_logging import console_logger

    logger = console_logger


paradex = Paradex(env=TESTNET, l1_address=TEST_L1_ADDRESS, l1_private_key=TEST_L1_PRIVATE_KEY)

recipient_address = ""
amount = Decimal(0)
logger.info(f"Transferring USDC from {TEST_L1_ADDRESS} to {recipient_address} amount {amount}")

asyncio.run(paradex.account.transfer_on_l2(recipient_address, amount))



"""
WebSocket message payload models based on AsyncAPI 2.6.0 specification.

These models represent the 'data' payload of WebSocket messages and are stricter
(required fields) than the REST API models to match the AsyncAPI spec.
"""

from typing import Any, Literal, cast

from pydantic import BaseModel, ConfigDict, Field


# Base WebSocket message structure
class WebSocketMessage(BaseModel):
    """Base WebSocket message with params and data."""

    model_config = ConfigDict(extra="allow")

    params: dict[str, Any] = Field(..., description="Message parameters including channel")
    data: dict[str, Any] = Field(..., description="Message payload data")


# WebSocket payload models (required fields per AsyncAPI spec)


class BalanceEventResponse(BaseModel):
    """Balance event WebSocket payload."""

    model_config = ConfigDict(extra="allow", populate_by_name=True)

    created_at: int = Field(..., description="Unix Millisecond timestamp at which the event occurred")
    fees: str = Field(..., description="Fees in USDC")
    fill_id: str = Field(..., description="Unique string ID for the fill")
    funding_index: str = Field(..., description="Funding index used for the fill")
    market: str = Field(..., description="Name of the market the update is referring to")
    realized_funding: str = Field(..., description="Fill's Realized Funding in USDC")
    realized_pnl: str = Field(..., description="Fill's realized PnL (excluding funding) in USDC")
    settlement_asset_balance_after: str = Field(..., description="Settlement asset balance after the fill")
    settlement_asset_balance_before: str = Field(..., description="Settlement asset balance before the fill")
    settlement_asset_price: str = Field(..., description="Settlement asset price")
    status: str = Field(..., description="Status of the fill")
    event_type: str = Field(alias="type", description="Type of the event")


class PriceBookUpdate(BaseModel):
    """Price book update WebSocket payload."""

    model_config = ConfigDict(extra="allow")

    best_ask_api: dict[str, Any] = Field(..., description="Size on the best ask from API (excluding RPI)")
    best_ask_interactive: dict[str, Any] = Field(..., description="Size on the best ask from UI")
    best_bid_api: dict[str, Any] = Field(..., description="Size on the best bid from API (excluding RPI)")
    best_bid_interactive: dict[str, Any] = Field(..., description="Size on the best bid from UI")
    deletes: list[dict[str, Any]] = Field(..., description="Deleted orders")
    inserts: list[dict[str, Any]] = Field(..., description="Inserted orders")
    last_updated_at: int = Field(..., description="Last update timestamp in milliseconds")
    market: str = Field(..., description="Market symbol")
    seq_no: int = Field(..., description="Sequence number")
    update_type: Literal["s", "d"] = Field(..., description="Update type")
    updates: list[dict[str, Any]] = Field(..., description="Updated orders")


# Type mapping for channel names to payload models
WS_PAYLOAD_MODELS = {
    "balance_events": BalanceEventResponse,
    "order_book": PriceBookUpdate,
}


def get_ws_payload_model(channel_name: str) -> type[BaseModel] | None:
    """Get the appropriate payload model for a WebSocket channel.

    Args:
        channel_name: The WebSocket channel name (e.g., "account", "bbo.BTC-USD-PERP")

    Returns:
        The corresponding payload model class, or None if not found
    """
    # Extract the base channel name (before any parameters)
    base_channel = channel_name.split(".")[0]
    return cast(type[BaseModel] | None, WS_PAYLOAD_MODELS.get(base_channel))


def validate_ws_payload(channel_name: str, payload: dict[str, Any]) -> BaseModel | None:
    """Validate a WebSocket payload against the appropriate model.

    Args:
        channel_name: The WebSocket channel name
        payload: The payload data to validate

    Returns:
        Validated model instance, or None if validation fails
    """
    model_class = get_ws_payload_model(channel_name)
    if model_class is None:
        return None

    try:
        return model_class.model_validate(payload)
    except Exception:
        return None



#!/usr/bin/env python3
"""
Example script demonstrating L2-only WebSocket usage with ParadexSubkey.

This example shows how to use WebSocket functionality with only L2 credentials (subkey mode),
without requiring L1 Ethereum address or private key.

Requirements:
- L2_PRIVATE_KEY: Starknet private key for the subkey
- L2_ADDRESS: L2 address of the main account (not the subkey address)

Usage:
    export L2_PRIVATE_KEY="0x..."
    export L2_ADDRESS="0x..."
    python examples/subkey_ws_api.py
"""

import asyncio
import os
import sys

from paradex_py import ParadexSubkey
from paradex_py.api.ws_client import ParadexWebsocketChannel
from paradex_py.environment import TESTNET

# Add the parent directory to the path so we can import paradex_py
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))


async def on_markets_summary(ws_channel, message):
    """Callback for markets summary WebSocket messages."""
    print(f"📊 Markets Summary: {message}")


async def on_orders(ws_channel, message):
    """Callback for orders WebSocket messages."""
    print(f"📋 Orders: {message}")


async def on_fills(ws_channel, message):
    """Callback for fills WebSocket messages."""
    print(f"💰 Fills: {message}")


async def main():
    """Main function demonstrating L2-only WebSocket usage."""
    # Get L2 credentials from environment variables
    l2_private_key = os.getenv("L2_PRIVATE_KEY")
    l2_address = os.getenv("L2_ADDRESS")

    if not l2_private_key:
        print("Error: L2_PRIVATE_KEY environment variable not set")
        print("Please set your L2 private key: export L2_PRIVATE_KEY='0x...'")
        return

    if not l2_address:
        print("Error: L2_ADDRESS environment variable not set")
        print("Please set the L2 address of the main account: export L2_ADDRESS='0x...'")
        return

    print("🚀 Starting Paradex L2-only WebSocket example...")
    print(f"L2 Address: {l2_address}")
    print(f"L2 Private Key: {l2_private_key[:10]}...")

    # Initialize ParadexSubkey with L2-only credentials
    paradex = ParadexSubkey(
        env=TESTNET,
        l2_private_key=l2_private_key,
        l2_address=l2_address,
    )

    try:
        print("\n🔌 Connecting to WebSocket...")
        await paradex.ws_client.connect()
        print("✅ WebSocket connected successfully!")

        # Subscribe to public channels
        print("\n📊 Subscribing to public channels...")
        await paradex.ws_client.subscribe(channel=ParadexWebsocketChannel.MARKETS_SUMMARY, callback=on_markets_summary)
        print("✅ Subscribed to markets summary")

        # Subscribe to private channels (requires authentication)
        print("\n🔐 Subscribing to private channels...")
        await paradex.ws_client.subscribe(
            channel=ParadexWebsocketChannel.ORDERS, callback=on_orders, params={"market": "ALL"}
        )
        print("✅ Subscribed to orders")

        await paradex.ws_client.subscribe(
            channel=ParadexWebsocketChannel.FILLS, callback=on_fills, params={"market": "ETH-USD-PERP"}
        )
        print("✅ Subscribed to fills")

        # Wait for messages (run forever like the original example)
        print("\n⏳ Listening for WebSocket messages...")
        print("(Press Ctrl+C to stop)")

        try:
            # Run forever like the original connect_ws_api.py
            await asyncio.Event().wait()
        except KeyboardInterrupt:
            print("\n🛑 Stopping...")

        print("\n🔌 WebSocket session completed!")

        print("\n🎉 L2-only WebSocket example completed successfully!")

    except Exception as e:
        print(f"❌ Error: {e}")
        import traceback

        traceback.print_exc()


if __name__ == "__main__":
    asyncio.run(main())



from decimal import Decimal
from enum import Enum
from typing import Any

from paradex_py.utils import time_now_milli_secs

decimal_zero = Decimal(0)


class OrderAction(Enum):
    NAN = "NAN"
    Send = "SEND"
    SendCancel = "SEND_CANCEL"


class OrderType(Enum):
    Market = "MARKET"
    Limit = "LIMIT"
    StopLimit = "STOP_LIMIT"
    StopMarket = "STOP_MARKET"
    TakeProfitLimit = "TAKE_PROFIT_LIMIT"
    TakeProfitMarket = "TAKE_PROFIT_MARKET"
    StopLossMarket = "STOP_LOSS_MARKET"
    StopLossLimit = "STOP_LOSS_LIMIT"


class OrderStatus(Enum):
    NEW = "NEW"
    OPEN = "OPEN"
    CLOSED = "CLOSED"


class OrderSide(Enum):
    Buy = "BUY"
    Sell = "SELL"

    def opposite_side(self) -> "OrderSide":
        return OrderSide.Sell if self == OrderSide.Buy else OrderSide.Buy

    def sign(self) -> int:
        return 1 if self == OrderSide.Buy else -1

    # 1 = buy, 2 = sell (for chain)
    def chain_side(self) -> str:
        return "1" if self == OrderSide.Buy else "2"


class Order:
    def __init__(
        self,
        market: str,
        order_type: OrderType,
        order_side: OrderSide,
        size: Decimal,
        limit_price: Decimal = decimal_zero,
        client_id: str = "",
        signature_timestamp: int | None = None,
        instruction: str = "GTC",
        reduce_only: bool = False,
        recv_window: int | None = None,
        stp: (
            str | None
        ) = None,  # Self Trade Prevention, EXPIRE_MAKER, EXPIRE_TAKER or EXPIRE_BOTH, default: EXPIRE_TAKER
        trigger_price: Decimal | None = None,
        order_id: str | None = None,
    ) -> None:
        ts = time_now_milli_secs()
        self.id = order_id
        self.account: str = ""
        self.status = OrderStatus.NEW
        self.limit_price = limit_price
        self.size = size
        self.market = market
        self.remaining = size
        self.order_type = order_type
        self.order_side = order_side
        self.client_id = client_id
        self.instruction = instruction
        self.reduce_only = reduce_only
        self.created_at = ts  # milliseconds
        self.cancel_reason = ""
        self.last_action = OrderAction.NAN
        self.last_action_time = 0
        self.cancel_attempts = 0
        self.signature = ""
        self.signature_timestamp = ts if signature_timestamp is None else signature_timestamp
        self.recv_window = recv_window
        self.stp = stp
        self.trigger_price = trigger_price

    def __repr__(self) -> str:
        ord_status = self.status.value
        if self.status == OrderStatus.CLOSED:
            ord_status += f"({self.cancel_reason})"
        msg = f"{self.market} {ord_status} {self.order_type.name} "
        msg += f"{self.order_side} {self.remaining}/{self.size}"
        msg += f"@{self.limit_price}" if self.is_limit_type() else ""
        msg += f"trigger@{self.trigger_price}" if self.trigger_price else ""
        msg += f"recv_window={self.recv_window}" if self.recv_window else ""
        msg += f";id={self.id}" if self.id else ""
        msg += f";client_id={self.client_id}" if self.client_id else ""
        msg += f";last_action:{self.last_action}" if self.last_action != OrderAction.NAN else ""
        msg += f";signed with:{self.signature}@{self.signature_timestamp}"
        return msg

    def __eq__(self, __o) -> bool:
        return self.id == __o.id

    def __hash__(self) -> int:
        return hash(self.id)

    def dump_to_dict(self) -> dict[Any, Any]:
        order_dict: dict[Any, Any] = {
            "market": self.market,
            "side": self.order_side.value,
            "size": str(self.size),
            "type": self.order_type.value,
            "client_id": self.client_id,
            "instruction": self.instruction,
            "signature": self.signature,
            "signature_timestamp": self.signature_timestamp,
            "recv_window": self.recv_window,
            "stp": self.stp,
        }
        if self.is_limit_type():
            order_dict["price"] = str(self.limit_price)
        if self.trigger_price:
            order_dict["trigger_price"] = str(self.trigger_price)
        if self.reduce_only:
            order_dict["flags"] = ["REDUCE_ONLY"]

        # For modify order
        if self.id:
            order_dict["id"] = self.id
        return order_dict

    def chain_price(self) -> str:
        if self.order_type == OrderType.Market:
            return "0"
        return str(int(self.limit_price.scaleb(8)))

    def chain_size(self) -> str:
        return str(int(self.size.scaleb(8)))

    def is_limit_type(self) -> bool:
        return self.order_type in [
            OrderType.Limit,
            OrderType.StopLimit,
            OrderType.TakeProfitLimit,
            OrderType.StopLossLimit,
        ]
